import hashlib
import json
import logging
import os
import traceback
import pathlib

from ansible.executor.task_result import TaskResult
from ansible.playbook.task import Task
from ansible.plugins.callback import CallbackBase

from ansible_bender.api import Application
from ansible_bender.builders.base import BuildState
from ansible_bender.constants import NO_CACHE_TAG

FILE_ACTIONS = ["file", "copy", "synchronize", "unarchive", "template"]
logger = logging.getLogger("ansible_bender")


class CallbackModule(CallbackBase):
    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'hard-worker'
    CALLBACK_NAME = 'a_container_image_snapshoter'
    CALLBACK_NEEDS_WHITELIST = True

    def _get_app_and_build(self):
        build_id = os.environ["AB_BUILD_ID"]
        db_path = os.environ["AB_DB_PATH"]
        app = Application(init_logging=False, db_path=db_path)
        build = app.get_build(build_id)
        app.set_logging(debug=build.debug, verbose=build.verbose)
        return app, build

    def _snapshot(self, task_result):
        """
        snapshot the target container

        :param task_result: instance of TaskResult
        """
        if task_result._task.action in ["setup", "gather_facts"]:
            # we ignore setup
            return
        if task_result.is_failed() or task_result._result.get("rc", 0) > 0:
            return
        a, build = self._get_app_and_build()
        if build.is_failed():
            return
        if "stop-layering" in getattr(task_result._task, "tags", []):
            build.stop_layering()
            a.db.record_build(build)
            self._display.display("detected tag 'stop-layering', tasks won't be cached nor layered any more")
            return
        if not build.is_layering_on():
            return
        content = self.get_task_content(task_result._task)
        if task_result.is_skipped() or getattr(task_result, "_result", {}).get("skip_reason", False):
            a.record_progress(None, content, None, build_id=build.build_id)
            return
        # # alternatively, we can guess it's a file action and do getattr(task, "src")
        # # most of the time ansible says changed=True even when the file is the same
        if task_result._task.action in FILE_ACTIONS:
            if not task_result.is_changed():
                status = a.maybe_load_from_cache(content, build_id=build.build_id)
                if status:
                    self._display.display("loaded from cache: '%s'" % status)
                    return
        image_name = a.cache_task_result(content, build)
        if image_name:
            self._display.display("caching the task result in an image '%s'" % image_name)

    @staticmethod
    def get_task_content(task: Task):
        sha512 = hashlib.sha512()
        serialized_data = task.get_ds()
        if not serialized_data:
            # ansible 2.8
            serialized_data = task.dump_attrs()
        if not serialized_data:
            logger.error("unable to obtain task content from ansible: caching will not work")
            return
        c = json.dumps(serialized_data, sort_keys=True)

        logger.debug("content = %s", c)
        sha512.update(c.encode("utf-8"))

        # If task is a file action, cache the src.
        #
        # Take the file stats of the src (if directory, get the stats
        # of every file within) and concatinate it with the task config
        # (assigned under serialized_data)
        #
        # The idea is that, if a file is changed, so will its modification time,
        # which will force the layer to be reloaded. Otherwise, load from cache.
        #
        # Note: serialized_data was grabbed above.
        task_config = task.dump_attrs()
        if( ('args' in task_config) and ('src' in task_config['args']) ):
            src = task_config['args']['src']
            src_path = os.path.join(task.get_search_path()[0], "files", src)

            if(not(os.path.exists(src_path))):
                src_path = os.path.join(task.get_search_path()[0], src)

            if os.path.isdir(src_path):
                dir_hash = CallbackModule.get_dir_fingerprint(src_path)
                sha512.update(dir_hash.encode("utf-8"))
            elif os.path.isfile(src_path):
                the_file = pathlib.Path(src_path)
                date_modified = str(the_file.stat().st_mtime)
                sha512.update(date_modified.encode("utf-8"))

        return sha512.hexdigest()

    @staticmethod
    def get_dir_fingerprint(directory):
        sha512 = hashlib.sha512()
        for root, dirs, files in os.walk(directory):
            for filename in files:
                the_file = pathlib.Path(os.path.join(directory, root, filename))

                if not the_file.exists():
                    continue

                date_modified = str(the_file.stat().st_mtime)
                sha512.update(date_modified.encode("utf-8"))

        return sha512.hexdigest()

    def _maybe_load_from_cache(self, task):
        """
        load image state from cache

        :param task: instance of Task
        """
        if task.action in ["setup", "gather_facts"]:
            # we ignore setup
            return
        a, build = self._get_app_and_build()
        if build.is_failed():
            # build failed, skip the task
            task.when = "0"  # skip
            return
        if "stop-layering" in getattr(task, "tags", []):
            build.stop_layering()
            a.db.record_build(build)
            return
        if NO_CACHE_TAG in getattr(task, "tags", []):
            self._display.display("detected tag '%s': won't load from cache from now" % NO_CACHE_TAG)
            build.cache_tasks = False
            a.db.record_build(build)
            return
        if not build.was_last_layer_cached():
            return
        if not build.is_layering_on():
            return
        content = self.get_task_content(task)
        logger.debug("hash = %s", content)
        status = a.maybe_load_from_cache(content, build_id=build.build_id)
        if status:
            self._display.display("loaded from cache: '%s'" % status)
            task.when = "0"  # skip

    def abort_build(self):
        logger.debug("%s", traceback.format_exc())
        a, build = self._get_app_and_build()
        a.db.record_build(build, build_state=BuildState.FAILED)

    def v2_playbook_on_task_start(self, task, is_conditional):
        try:
            return self._maybe_load_from_cache(task)
        except Exception as ex:
            logger.error("error while running the build: %s", ex)
            self.abort_build()

    def v2_on_any(self, *args, **kwargs):
        try:
            first_arg = args[0]
        except IndexError:
            return
        if isinstance(first_arg, TaskResult):
            try:
                return self._snapshot(first_arg)
            except Exception as ex:
                logger.error("error while running the build: %s", ex)
                self.abort_build()



"""
Base class for builders
"""
from enum import Enum


class BuildState(Enum):
    NEW = "new"
    IN_PROGRESS = "in_progress"
    DONE = "done"
    FAILED = "failed"


class Builder:
    ansible_connection = "default-value"
    name = "default-value"

    def __init__(self, build, debug=False):
        """
        :param build: instance of Build
        :param debug: bool, provide debug output if True
        """
        self.build = build
        self.ansible_host = None
        self.debug = debug
        self.python_interpr_prio = (
            "/usr/bin/python3",
            "/usr/local/bin/python3",
            "/usr/bin/python3.7",
            "/usr/bin/python37",
            "/usr/bin/python3.6",
            "/usr/bin/python36",
            "/usr/bin/python2",
            "/usr/local/bin/python2",
            "/usr/bin/python",
            "/usr/local/bin/python",
            "/usr/libexec/platform-python",
        )

    def create(self):
        """
        create a container where all the work happens
        """

    def run(self, image_name, command):
        """
        run provided command in the selected image and return output

        :param image_name: str
        :param command: list of str
        :return: str (output)
        """

    def commit(self, image_name):
        """
        snapshot the artifact and create an image

        :param image_name: str, name the snapshot
        """

    def clean(self):
        """
        clean working container
        """

    def get_image_id(self, image_name):
        """ return image_id for provided image """

    def is_image_present(self, image_reference):
        """
        :return: True when the selected image is present, False otherwise
        """

    def is_base_image_present(self):
        """
        :return: True when the base image is present, False otherwise
        """
        return self.is_image_present(self.build.base_image)

    def pull(self):
        """
        pull base image
        """

    def push(self, build, target, force=False):
        """
        push built image into a remote location

        :param target: str, transport:details
        :param build: instance of Build
        :param force: bool, bypass checks if True
        :return: None
        """

    def find_python_interpreter(self):
        """
        find python executable in the base image, for prio order see constructor

        :return: str, path to python interpreter
        """

    def get_logs(self):
        """
        obtain logs for the selected build

        :return: list of str
        """

    def sanity_check(self):
        """
        invoke container tooling and thus verify they work well
        """

    def check_container_creation(self):
        """
        check that containers can be created
        """



"""
Jsonschema for Build and ImageMetadata
"""

# thanks https://www.jsonschema.net/
BUILD_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "title": "Build Schema",
    "optional": [
        "build_id",
        "build_finished_time",
        "build_start_time",
        "build_volumes",
        "layers",
        "final_layer_id",
        "layer_index",
        "build_container",
        "cache_tasks",
        "log_lines",
        "layering",
        "squash",
        "debug",
        "verbose",
        "pulled",
        "ansible_extra_args",
        "python_interpreter"
    ],
    "required": [
        "playbook_path",
        "metadata",
        "state",
        "base_image",
        "target_image",
        "builder_name",
    ],
    "properties": {
        "build_id": {
            "type": ["null", "string"],
            "title": "Unique build identifier",
            "examples": [
                "1", "42", None
            ]
        },
        "playbook_path": {
            "type": "string",
            "title": "Path to playbook",
            "examples": [
                "./playbook.yaml"
            ]
        },
        "build_volumes": {
            "type": "array",
            "title": "List of volumes available during build",
            "items": {
                "type": "string",
                "examples": [
                    "/c:/d"
                ],
            }
        },
        "metadata": {
            "type": "object",
            "title": "Image metadata"
        },
        "state": {
            "type": "string",
            "title": "Build state",
            "examples": [
                "new"
            ],
        },
        "build_start_time": {
            "type": ["string", "null"],
            "title": "Time when the build started",
            "examples": [
                ""
            ]
        },
        "build_finished_time": {
            "type": ["string", "null"],
            "title": "Time when the build finished",
            "examples": [
                ""
            ]
        },
        "base_image": {
            "type": "string",
            "title": "Base image",
            "examples": [
                "fedora:29"
            ],
        },
        "target_image": {
            "type": "string",
            "title": "Name of the final image which we build",
            "examples": [
                "my-funky-image"
            ],
        },
        "builder_name": {
            "type": ["string", "null"],
            "title": "Builder to use during a build",
            "examples": [
                "buildah"
            ]
        },
        "layers": {
            "type": "array",
            "title": "Layers of the image"
        },
        "final_layer_id": {
            "type": ["null", "string"],
            "title": "ID of the final layer",
            "examples": [
                "...some hash..."
            ]
        },
        "layer_index": {
            "type": "object",
            "title": "index for layers"
        },
        "build_container": {
            "type": ["string", "null"],
            "title": "Name of the container where the build is happening",
            "examples": [
                "my-image-...-cont"
            ]
        },
        "cache_tasks": {
            "type": "boolean",
            "title": "Should we cache tasks?",
            "examples": [
                True
            ],
        },
        "log_lines": {
            "type": "array",
            "title": "A list of log lines"
        },
        "layering": {
            "type": "boolean",
            "title": "Should we layer after each task?",
            "examples": [
                False
            ],
        },
        "squash": {
            "type": "boolean",
            "title": "Should we squash the final image to a single layer?",
            "examples": [
                False
            ],
        },
        "debug": {
            "type": "boolean",
            "title": "Should we print debug logs?",
            "examples": [
                False
            ]
        },
        "verbose": {
            "type": "boolean",
            "title": "Should we act verbosely?",
            "examples": [
                True
            ]
        },
        "pulled": {
            "type": "boolean",
            "title": "Was the base image pulled?",
            "examples": [
                False
            ]
        },
        "ansible_extra_args": {
            "type": ["string", "null"],
            "title": "Extra args to pass to ansible-playbook",
            "examples": [
                "--vv -e key=value"
            ],
        },
        "python_interpreter": {
            "type": ["string", "null"],
            "title": "Path to python interpreter within the base image",
            "examples": [
                "/usr/bin/python4"
            ]
        }
    }
}

IMAGE_META_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "title": "Image Metadata Schema",
    "optional": [
        "working_dir",
        "labels",
        "annotations",
        "env_vars",
        "cmd",
        "entrypoint",
        "user",
        "ports",
        "volumes"
    ],
    "properties": {
        "working_dir": {
            "type": ["string", "null"],
            "title": "Path to a working directory within a container image",
            "examples": [
                "/workshop"
            ],
        },
        "labels": {
            "type": "object",
            "title": "Key/value data to apply to the final image",
            "additionalProperties": {"type": "string"}
        },
        "annotations": {
            "type": "object",
            "title": "Key/value data to apply to the final image",
            "additionalProperties": {"type": "string"}
        },
        "env_vars": {
            "type": "object",
            "title": "Implicit environment variables to set in a container",
            "additionalProperties": {"type": "string"}
        },
        "cmd": {
            "type": ["string", "null"],
            "title": "A default command to invoke the container",
            "default": "",
            "examples": [
                "command -x -y z"
            ],
        },
        "entrypoint": {
            "type": ["string", "null"],
            "title": "An entrypoint script to configure for the container",
            "examples": [
                "entrypoint.sh"
            ],
        },
        "user": {
            "$id": "#/properties/user",
            "type": ["string", "null"],
            "title": "UID or username used to invoke the container",
            "examples": [
                "leonardo"
            ],
        },
        "ports": {
            "type": "array",
            "title": "A list of ports which are meant to be exposed on the host",
            "items": {
                "type": "string",
                "examples": [
                    "80", "443"
                ],
            }
        },
        "volumes": {
            "type": "array",
            "title": "A list of paths which are meant to be hosted outside of the container",
            "items": {
                "type": "string",
                "examples": [
                    "/path/to/a/directory"
                ],
            }
        }
    },
    "additionalProperties": False
}

PLAYBOOK_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "title": "Playbook Schema",
    "additionalProperties": False,
    "properties": {
        "working_container": {
            "type": ["object", "null"],
            "title": "settings for the container where the build occurs",
            "additionalProperties": False,
            "properties": {
                "volumes": {
                    "type": "array",
                    "title": "A list of paths which are meant to be hosted outside of the container",
                    "items": {
                        "type": "string",
                        "examples": [
                            "/path/to/a/directory"
                        ],
                    }
                },
                "user": {
                    "type": ["string", "null"],
                    "title": "UID or username used to invoke the container",
                    "examples": [
                        "leonardo"
                    ],
                },
                "entrypoint": {
                    "type": ["string", "null"],
                    "title": "entrypoint used during build to prevent certain containers, with processes running forever, getting stuck.",
                    "examples": [
                        "echo"
                    ]
                }
            }
        },
        "target_image": {
            "type": ["object", "null"],
            "title": "metadata of the final image which we built",
            "additionalProperties": False,
            "properties": {
                "name": {
                    "type": ["string", "null"],
                    "title": "Path to a working directory within a container image",
                },
                "ports": {
                    "type": "array",
                    "title": "A list of ports which are meant to be exposed on the host",
                    "items": {
                        "type": "string",
                        "examples": [
                            "80", "443"
                        ],
                    }
                },
                "volumes": {
                    "type": "array",
                    "title": "A list of paths which are meant to be hosted outside of the container",
                    "items": {
                        "type": "string",
                        "examples": [
                            "/path/to/a/directory"
                        ],
                    }
                },
                "working_dir": {
                    "type": ["string", "null"],
                    "title": "Path to a working directory within a container image",
                    "examples": [
                        "/workshop"
                    ],
                },
                "labels": {
                    "type": "object",
                    "title": "Key/value data to apply to the final image",
                    "additionalProperties": {"type": "string"}
                },
                "annotations": {
                    "type": "object",
                    "title": "Key/value data to apply to the final image",
                    "additionalProperties": {"type": "string"}
                },
                "environment": {
                    "type": "object",
                    "title": "implicit environment variables to set in a container",
                    "additionalProperties": {"type": "string"}
                },
                "cmd": {
                    "type": ["string", "null"],
                    "title": "A default command to invoke the container",
                    "examples": [
                        "command -x -y z"
                    ],
                },
                "entrypoint": {
                    "type": ["string", "null"],
                    "title": "An entrypoint script to configure for the container",
                    "examples": [
                        "entrypoint.sh"
                    ],
                },
                "user": {
                    "type": ["string", "null"],
                    "title": "UID or username used to invoke the container",
                    "examples": [
                        "leonardo"
                    ],
                },
            },
        },
        "base_image": {
            "type": "string",
            "title": "name of the container image to use as a base",
            "examples": [
                "fedora:29"
            ],
        },
        "cache_tasks": {
            "type": "boolean",
            "title": "When true, enable caching mechanism"
        },
        "layering": {
            "type": "boolean",
            "title": "When true, snapshot the image after a task is executed",
        },
        "squash": {
            "type": "boolean",
            "title": "When true, squash the final image down to a single layer",
        },
        "verbose_layer_names": {
            "type": "boolean",
            "title": "tag layers with a verbose name if true (image-name + timestamp), defaults to false",
        },
        "ansible_extra_args": {
            "type": "string",
            "title": "provide extra arguments for ansible-playbook run",
        },
        "buildah_from_extra_args": {
            "type": "string",
            "title": "provide extra arguments for `buildah from` command"
        },
        "buildah_run_extra_args": {
            "type": "string",
            "title": "provide extra arguments for `buildah run` command",
        },
        "podman_run_extra_args": {
            "type": "string",
            "title": "provide extra arguments for `podman run` command",
        },
    },
}



"""
Utility functions. This module can't depend on anything within ab.
"""
import logging
import os
import random
import re
import shutil
import string
import subprocess
import threading

from ansible_bender.constants import OUT_LOGGER


logger = logging.getLogger(__name__)
out_logger = logging.getLogger(OUT_LOGGER)


def graceful_get(d, *keys, default=None):
    """
    recursively obtain value from nested dict

    :param d: dict
    :param keys:
    :param default: a value to return by default

    :return: value or None
    """
    response = d
    for k in keys:
        try:
            response = response[k]
        except (KeyError, AttributeError, TypeError) as ex:
            logger.debug("can't obtain %s: %s", k, ex)
            return default
    return response


class StreamLogger(threading.Thread):
    def __init__(self, stream, print_output=False, log_level=logging.DEBUG, log_output=True, buffer=None):
        super().__init__(daemon=True)  # the threads should not linger
        self.stream = stream
        self.buffer = buffer  # to easily share output, both stdout & stderr
        self.output = []
        self.log_level = log_level
        self.log_output = log_output
        self.print_output = print_output

    def run(self):
        for line in self.stream:
            line = line.rstrip("\n")
            if self.buffer is not None:
                self.buffer.append(line)
            self.output.append(line)
            if self.log_output:
                logger.log(self.log_level, line)
            if self.print_output:
                out_logger.info(line)

    def get_output(self):
        return "\n".join(self.output)


def run_cmd(cmd, return_output=False, ignore_status=False, print_output=False, log_stderr=True,
            save_output_in_exc=True,
            log_output=True, return_all_output=False, **kwargs):
    """
    run provided command on host system using the same user as you invoked this code, raises
    subprocess.CalledProcessError if it fails

    :param cmd: list of str
    :param return_output: bool, return output of the command
    :param return_all_output: bool, return output including stderr
    :param ignore_status: bool, do not fail in case nonzero return code
    :param kwargs: pass keyword arguments to subprocess.check_* functions; for more info,
            please check `help(subprocess.Popen)`
    :param print_output: bool, print output via print()
    :param log_stderr: bool, log errors to stdout as ERROR level
    :param log_output: bool, print output of the command to logs
    :param save_output_in_exc: bool, add command output to exception in case of an error
    :return: None or str
    """
    logger.info('running command: "%s"', cmd)
    logger.debug('%s', " ".join(cmd))  # so you can easily copy/pasta
    whole_output = []
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               universal_newlines=True, **kwargs)
    o = StreamLogger(process.stdout, print_output=print_output, log_level=logging.DEBUG,
                     log_output=log_output, buffer=whole_output)
    stderr_log_lvl = logging.ERROR if log_stderr else logging.DEBUG
    e = StreamLogger(process.stderr, print_output=print_output, log_level=stderr_log_lvl, buffer=whole_output)
    o.start()
    e.start()
    process.wait()
    o.join()
    e.join()

    if process.returncode > 0:
        if ignore_status:
            if return_output:
                return o.get_output()
            else:
                return process.returncode
        else:
            out = ""
            errout = ""
            if save_output_in_exc:
                out = "\n".join(whole_output)
                errout = e.get_output()
            raise subprocess.CalledProcessError(cmd=cmd, returncode=process.returncode,
                                                stderr=errout, output=out)
    if return_all_output:
        return whole_output
    if return_output:
        return o.get_output()


class CommandDoesNotExistException(Exception):
    pass


def env_get_or_fail_with(env_name, err_msg):
    try:
        return os.environ[env_name]
    except KeyError:
        raise RuntimeError(err_msg)


def one_of_commands_exists(commands, exc_msg) -> str:
    """
    Verify that the provided command exists. Raise CommandDoesNotExistException in case of an
    error or if the command does not exist.

    :param commands: str, command to check (python 3 only)
    :param exc_msg: str, message of exception when command does not exist
    :return: str, the command which exists
    """
    found = False
    for command in commands:
        found = bool(shutil.which(command))  # py3 only
        if found:
            return command
    if not found:
        raise CommandDoesNotExistException(exc_msg)


def ap_command_exists():
    return one_of_commands_exists(
        ["ansible-playbook-3", "ansible-playbook"],
        "ansible-playbook command doesn't seem to be available on your system. "
        "It is usually available in 'ansible' package or follow the upstream instructions "
        "available at https://docs.ansible.com/ansible/latest/installation_guide/"
        "intro_installation.html#installation-guide"
    )


def buildah_command_exists():
    return one_of_commands_exists(
        ["buildah"],
        "buildah command doesn't seem to be available on your system. "
        "Please follow the upstream instructions "
        "available at https://github.com/projectatomic/buildah/blob/master/install.md"
    )


def podman_command_exists():
    return one_of_commands_exists(
        ["podman"],
        "podman command doesn't seem to be available on your system. "
        "Please follow the upstream instructions "
        "available at https://github.com/containers/libpod/blob/master/install.md"
    )


def git_command_exists():
    return one_of_commands_exists(
        ["git"],
        "git command is not available"
    )


def git_clone_to_path(repo_url, path, ref="master"):
    """
    clone repo provided as repo_url into directory specified as path

    :param repo_url: str
    :param path: str
    :param ref: str, git ref
    """
    run_cmd(["git", "clone", "--depth", "1", "-b", ref, repo_url, path])


def set_logging(
        logger_name="ansible_bender",
        level=logging.INFO,
        handler_class=logging.StreamHandler,
        handler_kwargs=None,
        format='%(asctime)s.%(msecs).03d %(filename)-17s %(levelname)-6s %(message)s',
        date_format='%H:%M:%S'):
    """
    Set personal logger for this library.

    :param logger_name: str, name of the logger
    :param level: int, see logging.{DEBUG,INFO,ERROR,...}: level of logger and handler
    :param handler_class: logging.Handler instance, default is StreamHandler (/dev/stderr)
    :param handler_kwargs: dict, keyword arguments to handler's constructor
    :param format: str, formatting style
    :param date_format: str, date style in the logs
    :return: logger instance
    """
    logger = logging.getLogger(logger_name)
    # do we want to propagate to root logger?
    # logger.propagate = False
    logger.setLevel(level)

    # don't readd handler
    if not [x for x in logger.handlers if isinstance(x, handler_class)]:
        handler_kwargs = handler_kwargs or {}
        handler = handler_class(**handler_kwargs)
        handler.setLevel(level)
        formatter = logging.Formatter(format, date_format)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger


def random_str(size=10):
    """
    create random string of selected size

    :param size: int, length of the string
    :return: the string
    """
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(size))


def is_ansibles_python_2(ap_exe: str) -> bool:
    """
    Discover whether ansible-playbook is using python 2.

    :param ap_exe: path to the python executable

    :return: True if it's 2
    """
    out = run_cmd([ap_exe, "--version"], log_stderr=True, return_output=True)

    # python version = 3.7.2
    reg = r"python version = (\d)"

    reg_grp = re.findall(reg, out)
    try:
        py_version = reg_grp[0]
    except IndexError:
        logger.warning("could not figure out which python is %s using", ap_exe)
        return False  # we don't know, fingers crossed
    if py_version == "2":
        logger.info("%s is using python 2", ap_exe)
        return True
    logger.debug("it seems that %s is not using python 2", ap_exe)
    return False

def fancy_time(build_time):
    """ 
    returns build time in human readable form 

    :param build_time: total time taken for the build
    """
    seconds = build_time.seconds
    minutes = round(seconds / 60)
    hours = round(seconds / 60 / 60)
    days = build_time.days
    f_time = ""

    if days > 0:
        f_time = str(days)
        f_time = f_time + (" day" if days == 1 else " days")
    elif hours > 0:
        f_time = str(hours)
        f_time = f_time + (" hour" if hours == 1 else " hours")
    elif minutes > 0:
        f_time = str(minutes)
        f_time = f_time + (" minute" if minutes == 1 else " minutes")
    elif seconds > 0:
        f_time = str(seconds)
        f_time = f_time + (" second" if seconds == 1 else " seconds")
    else:
        f_time = "zero time"
    
    return f_time



import datetime
import logging
import os
import sys
from typing import Tuple, List

from ansible_bender.builder import get_builder
from ansible_bender.builders.base import BuildState
from ansible_bender.conf import Build
from ansible_bender.constants import OUT_LOGGER, OUT_LOGGER_FORMAT, TIMESTAMP_FORMAT
from ansible_bender.core import AnsibleRunner
from ansible_bender.db import Database
from ansible_bender.exceptions import ABBuildUnsuccesful
from ansible_bender.utils import set_logging


logger = logging.getLogger(__name__)
out_logger = logging.getLogger(OUT_LOGGER)


class Application:
    def __init__(self, debug=False, db_path=None, verbose=False, init_logging=True):
        """
        :param debug: bool, provide debug output if True
        :param db_path: str, path to json file where the database stores the data persistently
        :param verbose: bool, print verbose output
        :param init_logging: bool, set up logging if True
        """
        if init_logging:
            self.set_logging(debug=debug, verbose=verbose)
        self.verbose = verbose
        self.debug = debug
        self.db = Database(db_path=db_path)
        self.db_path = self.db.db_root_path

    @staticmethod
    def set_logging(debug: bool = False, verbose: bool = False):
        """ configure logging """
        if debug:
            set_logging(level=logging.DEBUG)
        elif verbose:
            set_logging(level=logging.INFO)
            set_logging(logger_name=OUT_LOGGER, level=logging.INFO, format=OUT_LOGGER_FORMAT,
                        handler_kwargs={"stream": sys.stdout})
        else:
            set_logging(level=logging.WARNING)
            set_logging(logger_name=OUT_LOGGER, level=logging.INFO, format=OUT_LOGGER_FORMAT,
                        handler_kwargs={"stream": sys.stdout})

    def build(self, build: Build):
        """
        build container image

        :param build: instance of Build
        """
        if not os.path.isfile(build.playbook_path):
            raise RuntimeError("No such file or directory: %s" % build.playbook_path)

        build.validate()
        build.metadata.validate()

        build.debug = self.debug
        build.verbose = self.verbose

        # we have to record as soon as possible
        self.db.record_build(build)

        try:
            builder = self.get_builder(build)
            builder.sanity_check()

            # before we start messing with the base image, we need to check for its presence first
            if not builder.is_base_image_present():
                builder.pull()
                build.pulled = True

            builder.check_container_creation()

            # let's record base image as a first layer
            base_image_id = builder.get_image_id(build.base_image)
            build.record_layer(None, base_image_id, None, cached=True)

            a_runner = AnsibleRunner(build.playbook_path, builder, build, debug=self.debug)

            # we are about to perform the build
            build.build_start_time = datetime.datetime.now()
            self.db.record_build(build, build_state=BuildState.IN_PROGRESS)

            build.python_interpreter = build.python_interpreter or self.db.load_python_interpreter(base_image_id)
            if not build.python_interpreter:
                build.python_interpreter = builder.find_python_interpreter()
                self.db.record_python_interpreter(base_image_id, build.python_interpreter)

            builder.create()
        except Exception:
            self.db.record_build(
                None,
                build_id=build.build_id,
                build_state=BuildState.FAILED,
                set_finish_time=True
            )
            raise

        try:
            try:
                output = a_runner.build(self.db_path)
            except ABBuildUnsuccesful as ex:
                b = self.db.record_build(None, build_id=build.build_id,
                                         build_state=BuildState.FAILED,
                                         set_finish_time=True)
                b.log_lines = ex.output.split("\n")
                self.db.record_build(b)
                timestamp = datetime.datetime.now().strftime(TIMESTAMP_FORMAT)
                image_name = build.target_image + "-" + timestamp + "-failed"
                b.target_image = image_name
                image_id = builder.commit(image_name)
                b.final_layer_id = image_id
                self.record_progress(b, None, image_id)
                out_logger.info("Image build failed /o\\")
                out_logger.info("The progress is saved into image '%s'", image_name)
                raise

            b = self.db.record_build(None, build_id=build.build_id, build_state=BuildState.DONE,
                                     set_finish_time=True)
            b.log_lines = output
            # commit the final image and apply all metadata
            b.final_layer_id = builder.commit(build.target_image, final_image=True)

            if b.squash:
                logger.debug("Squashing metadata into a single layer")
                # reset layers if squashing
                b.wipe_layers()
                self.record_progress(b, None, b.final_layer_id)
            if not b.is_layering_on():
                self.record_progress(b, None, b.final_layer_id)
            else:
                self.db.record_build(b)

            out_logger.info("Image '%s' was built successfully \\o/",  build.target_image)
        finally:
            builder.clean()

    def get_build(self, build_id: str = None) -> Build:
        """
        get selected build or latest build if build_id is None

        :param build_id: str or None
        :return: build
        """
        if build_id is None:
            return self.db.get_latest_build()
        return self.db.get_build(build_id)

    def get_logs(self, build_id: str = None) -> List[str]:
        """
        get logs for a specific build, if build_id is not, select the latest build

        :param build_id: str or None
        :return: list of str
        """
        build = self.get_build(build_id=build_id)
        return build.log_lines

    def list_builds(self) -> List[Build]:
        return self.db.load_builds()

    def inspect(self, build_id: str = None):
        """
        provide detailed information about the selected build

        :param build_id: str or None
        :return: dict
        """
        build = self.get_build(build_id=build_id)
        di = build.to_dict()
        del di["log_lines"]  # we have a dedicated command for that
        del di["layer_index"]  # internal info
        return di

    def push(self, target, build_id: str = None, force: bool = False):
        """
        push built image into a remote location, this method raises an exception when:
         * the push failed or the image can't be found
         * the build haven't finished yet

        :param target: str, transport:details
        :param build_id: id of the build or None
        :param force: bool, bypass checks if True
        :return: None
        """
        build = self.get_build(build_id=build_id)
        builder = self.get_builder(build)
        builder.push(build, target, force=force)

    def get_builder(self, build: Build):
        return get_builder(build.builder_name)(build, debug=self.debug)

    def maybe_load_from_cache(self, content: str, build_id: str) -> str:
        if not content:
            return

        build = self.db.get_build(build_id)
        builder = self.get_builder(build)

        if not build.cache_tasks:
            return

        base_image_id, layer_id = self.record_progress(build, content, None)
        builder.swap_working_container()
        return layer_id

    def get_layer(self, content: str, base_image_id: str) -> str:
        """
        provide a layer for given content and base_image_id; if there
        is such layer in cache store, return its layer_id

        :param content:
        :param base_image_id:
        :return:
        """
        return self.db.get_cached_layer(content, base_image_id)

    def record_progress(self, build: Build, content: str, layer_id: str, build_id: str = None) -> Tuple[str, str]:
        """
        record build progress to the database

        :param build:
        :param content: str or None
        :param layer_id:
        :param build_id:
        :return:
        """
        if build_id:
            build = self.db.get_build(build_id)
        base_image_id = build.get_top_layer_id()
        was_cached = False
        if not layer_id:
            # skipped task, it was cached
            if content:
                layer_id = self.get_layer(content, base_image_id)
                builder = self.get_builder(build)
                if not builder.is_image_present(layer_id):
                    logger.info("layer %s for content %s does not exist", layer_id, content)
                    layer_id = None
            if not layer_id:
                return None, None
            was_cached = True
        build.record_layer(content, layer_id, base_image_id, cached=was_cached)
        self.db.record_build(build)
        return base_image_id, layer_id

    def create_new_layer(self, content: str, build: Build) -> Tuple[str, str, str]:
        """
        create new layer from the current state of the container of specified build

        :param content: task as a str
        :param build: Build instance
        :return:
        """
        builder = self.get_builder(build)
        image_name = ""
        if build.verbose_layer_names:
            timestamp = datetime.datetime.now().strftime(TIMESTAMP_FORMAT)
            image_name = "%s-%s" % (build.target_image, timestamp)
            # buildah doesn't accept upper case
            image_name = image_name.lower()
        layer_id = builder.commit(image_name, print_output=False)
        base_image_id, _ = self.record_progress(build, content, layer_id)
        return image_name, layer_id, base_image_id

    def cache_task_result(self, content: str, build: Build) -> str:
        """ snapshot the container after a task was executed """
        if not content:
            logger.info("no content provided, will not cache this layer")
            return
        image_name, layer_id, base_image_id = self.create_new_layer(content, build)
        if not build.cache_tasks:  # actually we could still cache results
            return
        self.db.save_layer(layer_id, base_image_id, content)
        return image_name

    def clean(self):
        self.db.release()

    def remove_build(self, build_id: str):
        self.db.delete_build(build_id)



class ABError(Exception):
    pass


class ABValidationError(ABError):
    pass


class ABBuildUnsuccesful(ABError):
    """ Build was not successful """
    def __init__(self, msg, output):
        self.msg = msg
        self.output = output

    def __str__(self):
        return "%s" % self.msg



""" Build inside OKD as a custom build """
import json
import os
import shutil
import tempfile

from ansible_bender.builders.base import BuildState
from ansible_bender.conf import ImageMetadata, Build
from ansible_bender.utils import env_get_or_fail_with, graceful_get, git_clone_to_path


def okd_load_metadata():
    """ load metadata about the build from the BUILD env var """
    b = env_get_or_fail_with("BUILD", "BUILD environment variable is not set, are you running in openshift?")
    bd = json.loads(b)
    response = (
        graceful_get(bd, "spec", "source", "git", "uri"),
        graceful_get(bd, "spec", "source", "git", "ref"),
        graceful_get(bd, "spec", "output", "to", "name"),
    )
    if not all(response):
        raise RuntimeError("Not all build parameters seem to be set, halting.")
    return response


def okd_get_playbook_base():
    """ load metadata from os.environ and return playbook path & base image name """
    return (env_get_or_fail_with("AB_PLAYBOOK_PATH", "Can't get playbook path from the environment"),
            env_get_or_fail_with("AB_BASE_IMAGE", "Can't get base image name from the environment"))


def build_inside_openshift(app):
    """
    This is expected to run inside an openshift pod spawned via custom build

    :param app: instance of Application
    """
    playbook_path, base_image = okd_get_playbook_base()

    if playbook_path.startswith("/"):
        raise RuntimeError("The path to playbook needs to be relative within the git repo.")

    uri, ref, target_image = okd_load_metadata()

    tmp = tempfile.mkdtemp(prefix="ab-okd")

    try:
        git_clone_to_path(uri, tmp, ref=ref)

        playbook_path = os.path.abspath(os.path.join(tmp, playbook_path))
        if not playbook_path.startswith(tmp):
            raise RuntimeError("The path to playbook points outside of the git repo, this is not allowed.")

        build = Build()
        build.metadata = ImageMetadata()  # TODO: needs to be figured out
        build.playbook_path = playbook_path
        build.base_image = base_image
        build.target_image = target_image
        build.builder_name = "buildah"
        build.cache_tasks = False  # we have local storage in pod, so this doesn't make any sense
        app.build(build)

    finally:
        shutil.rmtree(tmp)
    # TODO: push



OUT_LOGGER = "ab-out"
OUT_LOGGER_FORMAT = "%(message)s"
TIMESTAMP_FORMAT = "%Y%m%d-%H%M%S%f"
TIMESTAMP_FORMAT_TOGETHER = "%Y%m%d%H%M%S%f"
NO_CACHE_TAG = "no-cache"

# configuration related constants
ANNOTATIONS_KEY = "annotations"

# ansible playbook template in yaml
PLAYBOOK_TEMPLATE = """---
- name: Containerized version of $project
  hosts: all
  vars:
    a_variable: value
    # configuration specific for ansible-bender
    ansible_bender:
      base_image: fedora:latest
      target_image:
        # command to run by default when invoking the container
        cmd: /command.sh
        name: $project
      working_container:
        volumes:
        # mount this git repo to the working container at /src
        - "{{ playbook_dir }}:/src"
  tasks:
  - name: install dependencies needed to run project $project
    package:
      name:
      - a_package
      - another_package
      state: present
"""



"""
Module to interact with Ansible, perform ansible-playbook and extract metadata from Ansible vars

A sample configuration:

- hosts: all
  vars:
    ansible_bender:

      base_image: fedora:28

      target_image:
        name: asdqwe
        # configure environment variables: same name as ansible
        environment:
          X: Y
        labels:
          key: value
        # default working directory
        working_dir: /path

      working_container:
        user: 12345  # TODO
        volumes:
        - { src: /path/to/my/code, dest: /src:Z }

tasks:
  ...

"""
import copy
import datetime
import importlib
import json
import logging
import os
import shlex
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path
import uuid
import configparser

import jsonschema
import yaml

import ansible_bender
from ansible_bender import callback_plugins
from ansible_bender.conf import ImageMetadata, Build
from ansible_bender.constants import TIMESTAMP_FORMAT, TIMESTAMP_FORMAT_TOGETHER
from ansible_bender.exceptions import ABBuildUnsuccesful, ABValidationError
from ansible_bender.schema import PLAYBOOK_SCHEMA
from ansible_bender.utils import run_cmd, ap_command_exists, random_str, graceful_get, \
    is_ansibles_python_2

from ansible.inventory.manager import InventoryManager
from ansible.vars.manager import VariableManager
from ansible.parsing.dataloader import DataLoader
from ansible_bender.conf import Build, ImageMetadata

logger = logging.getLogger(__name__)
# callback_whitelist got renamed to callbacks_enabled in ansible
# drop callback_whitelist once 2.15 is released
# https://docs.ansible.com/ansible/latest/reference_appendices/config.html#callbacks-enabled
A_CFG_TEMPLATE = """\
[defaults]
# when user is changed, ansible might not be able to write to /.ansible
remote_tmp = /tmp
callback_plugins={0}
callback_whitelist=snapshoter\n
callbacks_enabled=snapshoter\n
"""


def run_playbook(playbook_path, inventory_path, a_cfg_path, connection, extra_variables=None,
                 ansible_args=None, debug=False, environment=None, try_unshare=True,
                 provide_output=True, log_stderr=False):
    """
    run selected ansible playbook and return output from ansible-playbook run

    :param playbook_path:
    :param inventory_path:
    :param a_cfg_path:
    :param connection:
    :param extra_variables:
    :param ansible_args: list of str, extra arguments for a-p
    :param debug:
    :param environment:
    :param try_unshare: bool, do `buildah unshare` if uid > 0
    :param provide_output: bool, present output to user
    :param log_stderr: bool, log errors coming from stderr to our logger

    :return: output
    """
    ap = ap_command_exists()
    if is_ansibles_python_2(ap):
        # I just realized it could work, we would just had to disable the
        # callback plugin: no caching and layering
        raise RuntimeError(
            "ansible-bender is written in python 3 and does not work in python 2,\n"
            f"it seems that {ap} is using python 2 - ansible-bender will not"
            "work in such environment\n"
        )
    cmd_args = [
        ap,
        "-c", connection,

    ]
    if inventory_path:
        cmd_args += ["-i", inventory_path]
    if debug:
        cmd_args += ["-vvv"]
    if extra_variables:
        cmd_args += ["--extra-vars"] + \
                    [" ".join(
                        ["{}={}".format(k, v)
                         for k, v in extra_variables.items()]
                    )]
    if ansible_args:
        cmd_args += ansible_args
    cmd_args += [playbook_path]
    logger.debug("%s", " ".join(cmd_args))

    env = os.environ.copy()
    env["ANSIBLE_RETRY_FILES_ENABLED"] = "0"
    if debug:
        env["ANSIBLE_STDOUT_CALLBACK"] = "debug"
    if environment:
        env.update(environment)
    if a_cfg_path:
        env["ANSIBLE_CONFIG"] = a_cfg_path

    if try_unshare and os.getuid() != 0:
        logger.info("we are running rootless, prepending `buildah unshare`")
        # rootless, we need to `buildah unshare` for sake of `buildah mount`
        # https://github.com/containers/buildah/issues/1271
        # the need for `--` https://github.com/containers/buildah/issues/1374
        cmd_args = ["buildah", "unshare", "--"] + cmd_args

    # ansible has no official python API, the API they have is internal and said to break compat
    try:
        return run_cmd(
            cmd_args,
            print_output=provide_output,
            save_output_in_exc=True,
            env=env,
            return_all_output=provide_output,
            log_stderr=log_stderr,
        )
    except subprocess.CalledProcessError as ex:
        raise ABBuildUnsuccesful("ansible-playbook execution failed: %s" % ex, ex.output)


class AnsibleRunner:
    """
    Run ansible on provided artifact using the Builder interface
    """

    def __init__(self, playbook_path, builder, build, debug=False):
        self.build_i = build
        self.pb = playbook_path
        self.builder = builder
        self.debug = debug

    def _create_inventory_file(self, fd, python_interpreter):
        fd.write(
            '%s ansible_connection="%s" ansible_python_interpreter="%s"\n' % (
                self.builder.ansible_host,
                self.builder.ansible_connection,
                python_interpreter
            )
        )

    def _create_ansible_cfg(self, fd):
        callback_plugins_dir = os.path.dirname(callback_plugins.__file__)
        fd.write(A_CFG_TEMPLATE.format(callback_plugins_dir))

    def _get_path_our_site(self):
        """ return a path to a directory which contains ansible_bender installation """
        # pip in Fedora installs to /usr/local which is on default pythonpath
        # but when ansible invokes the callback plugin, that directory is not on sys.path: wat?
        # hence, let's add the site ab is installed in to sys.path
        return os.path.dirname(os.path.dirname(ansible_bender.__file__))

    def _correct_host_entries(self, playbook_path, tmpDir):
        """ Correct the host entries in the playbook and all imported playbooks """
        tmp_pb_path = os.path.join(tmpDir, "ab_" + str(uuid.uuid4()) + ".yaml")

        with open(playbook_path, "r") as fd_r:
            pb_dict = yaml.safe_load(fd_r)
            
        for idx, doc in enumerate(pb_dict):
            imported_playbook = doc.get("import_playbook")
            if imported_playbook:
                import_base_path = os.path.dirname(playbook_path)
                imported_playbook_path = os.path.join(import_base_path, imported_playbook)
                logger.debug("Encountered import_playbook, correcting hosts entries in imported file: %s", imported_playbook_path)
                doc["import_playbook"] = self._correct_host_entries(imported_playbook_path, tmpDir)
            else:
                host = doc["hosts"]
                logger.debug("play[%s], host = %s", idx, host)
                doc["hosts"] = self.builder.ansible_host

        with open(tmp_pb_path, "w") as fd:
            yaml.safe_dump(pb_dict, fd)

        return tmp_pb_path
    
    def build(self, db_path):
        """
        run the playbook against the container

        :param db_path, str, path to ab's database

        :return: str, output
        """
        tmp = tempfile.mkdtemp(prefix="ab")

        pythonpath = os.environ.get("PYTHONPATH", "")
        site_path = self._get_path_our_site()
        if pythonpath:
            pythonpath = f"{site_path}:{pythonpath}"
        else:
            pythonpath = site_path

        try:
            environment = {
                "AB_BUILD_ID": self.build_i.build_id,
                "AB_DB_PATH": db_path,
                "PYTHONPATH": pythonpath,  # TODO write an e2e test for this
            }
            inv_path = os.path.join(tmp, "inventory")
            logger.info("creating inventory file %s", inv_path)
            with open(inv_path, "w") as fd:
                self._create_inventory_file(fd, self.build_i.python_interpreter)
            a_cfg_path = os.path.join(tmp, "ansible.cfg")
            with open(a_cfg_path, "w") as fd:
                self._create_ansible_cfg(fd)

            tmp_pb_path = self._correct_host_entries(self.pb, tmp)
            playbook_base = os.path.basename(self.pb).split(".", 1)[0]
            timestamp = datetime.datetime.now().strftime(TIMESTAMP_FORMAT)
            symlink_name = f".{playbook_base}-{timestamp}-{random_str()}.yaml"
            playbook_dir = os.path.dirname(self.pb)
            symlink_path = os.path.join(playbook_dir, symlink_name)
            os.symlink(tmp_pb_path, symlink_path)

            extra_args = None
            try:
                if self.build_i.ansible_extra_args:
                    extra_args = shlex.split(self.build_i.ansible_extra_args)
                return run_playbook(
                    symlink_path, inv_path, a_cfg_path, self.builder.ansible_connection,
                    debug=self.debug, environment=environment, ansible_args=extra_args
                )
            finally:
                os.unlink(symlink_path)
        finally:
            shutil.rmtree(tmp)


class PbVarsParser:
    def __init__(self, playbook_path):
        """
        :param playbook_path: str, path to playbook
        """
        self.playbook_path = playbook_path
        self.build = Build()
        self.metadata = ImageMetadata()
        self.build.metadata = self.metadata

    def expand_pb_vars(self):
        """
        populate vars from a playbook, defined in vars section

        :return: dict with the content of ansible_bender var
        """
        with open(self.playbook_path) as fd:
            plays = yaml.safe_load(fd)

        for play in plays[1:]:
            bender_vars = graceful_get(play, "vars", "ansible_bender")
            if bender_vars:
                logger.warning("Variables are loaded only from the first play.")

        try:
            # we care only about the first play, we don't want to merge dicts
            d = plays[0]
        except IndexError:
            raise RuntimeError("Invalid playbook, can't access the first document.")

        bender_vars = graceful_get(d, "vars", "ansible_bender")
        if not bender_vars:
            logger.info("no bender data found in the playbook")
            return {}

        tmp = tempfile.mkdtemp(prefix="ab")
        json_data_path = os.path.join(tmp, "j.json")

        # we cannot use "vars" variable because the variables are not expanded in there
        pb_vars = copy.deepcopy(d["vars"])
        while True:
            # just in case the variable is already defined
            timestamp = datetime.datetime.now().strftime(TIMESTAMP_FORMAT_TOGETHER)
            ab_vars_key = f"ab_vars_{timestamp}"
            if ab_vars_key not in pb_vars:
                logger.debug("ab vars key = %s", ab_vars_key)
                pb_vars[ab_vars_key] = d["vars"]
                break
        jinja_pb_vars_key = '{{ %s }}' % ab_vars_key
        pb = {
            "name": "Let Ansible expand variables",
            "hosts": "localhost",
            "vars": pb_vars,
            "vars_files": d.get("vars_files", []),
            "gather_facts": False,
            "tasks": [
                {"debug": {"msg": jinja_pb_vars_key}},
                {
                    "copy": {
                        "dest": json_data_path,
                        "content": jinja_pb_vars_key
                    }
                }
            ]
        }

        i_path = os.path.join(tmp, "i")
        with open(i_path, "w") as fd:
            fd.write("localhost ansible_connection=local")

        # json is easier to parse than yaml
        tmp_pb_path = os.path.join(tmp, "p.json")
        with open(tmp_pb_path, "w") as fd:
            json.dump([pb], fd)

        playbook_base = os.path.basename(self.playbook_path).split(".", 1)[0]
        symlink_name = f".{playbook_base}-{timestamp}-{random_str()}.yaml"
        playbook_dir = os.path.dirname(self.playbook_path)
        symlink_path = os.path.join(playbook_dir, symlink_name)
        os.symlink(tmp_pb_path, symlink_path)

        # yeah, ansible is not very smart for connection=local
        args = ["-e", f"ansible_python_interpreter={sys.executable}"]

        try:
            run_playbook(symlink_path, i_path, None, connection="local", try_unshare=False,
                         provide_output=False, log_stderr=True, ansible_args=args)

            with open(json_data_path) as fd:
                return json.load(fd)["ansible_bender"]
        finally:
            os.unlink(symlink_path)
            shutil.rmtree(tmp)

    def process_pb_vars(self, bender_data):
        """
        accept variables from the playbook and update the Build and ImageMetadata objects with them

        :param bender_data: dict with the content of ansible_bender vars
        :return:
        """
        if not bender_data:
            return
        try:
            # validation to error out unknown keys in /vars/ansible_bender
            jsonschema.validate(bender_data, PLAYBOOK_SCHEMA)
        except jsonschema.ValidationError as validation_error:
            if validation_error.validator == "type":
                # error is due to invalid value datatype
                path = "/" + "/".join(validation_error.path)
                expected_types = validation_error.validator_value
                if isinstance(validation_error.validator_value, list):
                    expected_types = ", ".join(validation_error.validator_value)
                message = f"variable {path} is set to {validation_error.instance}" \
                          f", which is not of type {expected_types}"
                raise ABValidationError(message) from validation_error
            else:
                # error is due to absence of a required key, unknown keys playbook or any other kind
                raise ABValidationError(validation_error.message) from validation_error

        self.metadata.update_from_configuration(bender_data.get("target_image", {}))
        self.build.update_from_configuration(bender_data)

    def get_build_and_metadata(self):
        """
        extra vars from the selected playbook

        :return: Build(), ImageMetadata()
        """
        bender_data = self.expand_pb_vars()

        self.process_pb_vars(bender_data)

        return self.build, self.metadata


class AnsibleVarsParser:
    """
    AnsibleVarParser is used to parse the ansible variables from the playbook and inventory
    This class attempts to use the ansible API to parse the variables
    If the inventory is not specified, the class attempts to find the inventory file
    If the inventory is not found, the class will fall back to PbVarsParser to parse the variables
    """

    # Prefix for ansible_bender variables
    _variable_prefix = "ansible_bender"

    _inventory_path = None

    _playbook_path = None

    _loader: DataLoader = None

    _inventory: InventoryManager = None

    _variable_manager: VariableManager = None

    _hosts_in_playbook: list = None

    _hosts: dict = {}

    build: Build = None

    metadata: ImageMetadata = None

    def __init__(self, playbook_path, inventory_path = None) -> None:
        self._base_dir = os.path.dirname(playbook_path)
        self._playbook_path = playbook_path
        self._inventory_path = inventory_path if inventory_path else self._find_inventory_path()
        self.build = Build()
        self.metadata = ImageMetadata()
        self.build.metadata = self.metadata

    def __getitem__(self, key):
        return self._hosts[key]

    def _find_ansible_cfg(self) -> str:
        """
        Find the ansible.cfg file
        Precendence: https://docs.ansible.com/ansible/latest/reference_appendices/config.html
        """
        # Check environment variable
        if "ANSIBLE_CONFIG" in os.environ:
            return os.environ["ANSIBLE_CONFIG"]

        # Check in self._base_dir (playbook directory)
        if os.path.isfile(os.path.join(self._base_dir, "ansible.cfg")):
            return os.path.join(self._base_dir, "ansible.cfg")

        # Check in home directory
        if os.path.isfile(os.path.join(os.path.expanduser("~"), ".ansible.cfg")):
            return os.path.join(os.path.expanduser("~"), ".ansible.cfg")

        # Check in /etc/ansible/ansible.cfg
        if os.path.isfile("/etc/ansible/ansible.cfg"):
            return "/etc/ansible/ansible.cfg"

        # No ansible.cfg file found
        return None

    def _find_inventory_path(self) -> str:
        """
        Get the inventory path from either
        1. ansible.cfg
        2. the default ansible inventory location (/etc/ansible/hosts)
        """
        ansible_cfg_path = self._find_ansible_cfg()

        if ansible_cfg_path:
            return self._get_inventory_from_ansible_cfg(ansible_cfg_path)
        else:
            if os.path.isfile("/etc/ansible/hosts"):
                return"/etc/ansible/hosts"

        return None

    def _get_inventory_from_ansible_cfg(self, ansible_cfg_path) -> str:
        """
        Get the inventory path from the ansible.cfg file
        """
        config = configparser.ConfigParser()
        config.read(ansible_cfg_path)
        inv_path = config.get("defaults", "inventory", fallback=None)

        # Possible to have an ansible.cfg file without an inventory path
        if inv_path is None:
            return None

        # Inventory path can be absolute or relative in ansible.cfg
        if os.path.isabs(inv_path):
            return inv_path
        else:
            return os.path.join(os.path.dirname(ansible_cfg_path), inv_path)

    def _get_host_names(self, hosts) -> list:
        """
        Parse the hosts from the playbook
        """
        parsed_host_information = []

        for host in hosts:
            parsed_host_information.append(host.get("hosts"))

        return parsed_host_information

    def _extract_variables(self, ansible_variables, variables):
        """
        Convinience function to get the variables from the variables dictionary
        Gets the ansible_bender variables and the prepended variables
        """

        if variables is None:
            return ansible_variables

        for key, value in variables.items():
            if str(key).startswith(self._variable_prefix):
                    # Assume that ansible API has already 
                    # pulled variables with correct precedence
                new_key = key.replace(self._variable_prefix, "")

                    # key can be prepended with ansible_bender_ or ansible_bender
                    # if it is prepended with ansible_bender_, remove the _
                if new_key.startswith("_"):
                    new_key = new_key[1:]

                    # If the new_key is blank
                    # then the "ansible_bender" variable was used
                    # Add the variables to the ansible_variables
                if new_key == "":
                    ansible_variables.update(value)
                else:
                    ansible_variables[new_key] = value

        return ansible_variables

    def _get_vars_from_playbook(self, host_name) -> dict:
        """
        Parse the playbook variables related to the ansible_bender information
        """
        playbook_variables = {}

        try:
            for host in self._hosts_in_playbook:
                if host.get("hosts") == host_name:
                    playbook_variables = self._extract_variables(playbook_variables, host.get("vars"))
        except Exception as e:
            print("Error getting playbook variables for host: " + host_name + " - " + str(e))
            pass

        return playbook_variables

    def _get_vars_for_host(self, host_name) -> dict:
        """
        Parse the host variables from the inventory information
        """
        ansible_variables = {}
        try:
            ansible_host = self._inventory.get_host(host_name)
            all_vars_for_host = self._variable_manager.get_vars(host=ansible_host, 
                                                            include_hostvars=True, 
                                                            include_delegate_to=True)

            # find prepended variables
            self._extract_variables(ansible_variables, all_vars_for_host)                
        except Exception as e:
            print("Error getting host and group vars for host: " + host_name + " - " + str(e))
            pass

        # Append the playbook variables to the ansible variables
        ansible_variables.update(self._get_vars_from_playbook(host_name))

        return ansible_variables

    
    def get_build_and_metadata(self):
        """
        Find all hosts in the playbook and get their variables
        return the build and imagemetadata information
        """
        if self._inventory_path:
            # Ansible API objects used to pull variables
            relative_playbook_file_path = os.path.basename(self._playbook_path)
            self._loader = DataLoader()
            self._loader.set_basedir(self._base_dir)
            self._inventory = InventoryManager(loader=self._loader, sources=self._inventory_path)
            self._variable_manager = VariableManager(loader=self._loader, inventory=self._inventory)
            self._hosts_in_playbook = self._loader.load_from_file(relative_playbook_file_path)

            host_names = self._get_host_names(self._hosts_in_playbook)
            for host_name in host_names:
                ansible_vars = self._get_vars_for_host(host_name)
                self._hosts[host_name] = ansible_vars

            # We only support one host in the playbook
            host = self._hosts[list(self._hosts)[0]] if len(self._hosts) > 0 else None

            if not host:
                raise Exception("No hosts found in the playbook")

            self.metadata.update_from_configuration(host.get("target_image", {}))
            self.build.update_from_configuration(host)
        else:
            # At this point
            # 1. No inventory was provided
            # 2. No ansible.cfg file was found with inventory information
            # 3. No default inventory file was found
            # So forget about using the ansible API to get variables
            # Just manually parse the playbook
            pb_vars_p = PbVarsParser(self._playbook_path)
            self.build, self.metadata = pb_vars_p.get_build_and_metadata()

        return self.build, self.metadata



"""
CLI for ansible builder
"""

import argparse
import json
import sys
import subprocess
import pkg_resources
import yaml
from tabulate import tabulate

from ansible_bender.api import Application
from ansible_bender.constants import ANNOTATIONS_KEY, PLAYBOOK_TEMPLATE
from ansible_bender.core import AnsibleVarsParser
from ansible_bender.db import PATH_CANDIDATES
from ansible_bender.okd import build_inside_openshift

from ansible_bender.utils import fancy_time, run_cmd


def split_once_or_fail_with(strink, pattern, error_message):
    """
    split selected string (string is a builtin, hence strink) using the given pattern;
    raise RuntimeError with provided error message if that fails

    :param strink: str, string to split
    :param pattern: str, pattern to split on
    :param error_message: str, error message to raise in case of split fails
    :return: (str, str)
    """
    try:
        k, v = strink.split(pattern, 1)
    except ValueError:
        raise RuntimeError(error_message)
    # we do this on purpose, so that ValueError is raised here
    return k, v


class CLI:
    def __init__(self):
        self.parser = argparse.ArgumentParser(
            prog='ansible-bender',
            description='Ansible builder = ansible-playbook + {buildah,docker}  '
                        '# create your container images with Ansible! ',
            epilog="Please use '--' to separate options and arguments."
        )
        self.parser.add_argument("-v", "--verbose", action="store_true",
                                 help="provide verbose output")
        self.parser.add_argument("--debug", action="store_true",
                                 help="provide all the output")
        self.parser.add_argument("-V", "--version", action="store_true",
                                 help="print version of ansible-bender")

        candidates_str = ", ".join(filter(lambda x: x, PATH_CANDIDATES))
        self.parser.add_argument(
            "--database-dir",
            action="store",
            help="a path to directory where ab will store runtime data, defaults to: \"%s\""
                 % candidates_str
        )
        self.subparsers = self.parser.add_subparsers()

        self._do_build_interface()
        self._do_list_builds_interface()
        self._do_get_logs_interface()
        self._do_inspect_interface()
        self._do_push_interface()
        self._do_clean_interface()
        self._do_init_interface()

        self.args = self.parser.parse_args()
        debug = False
        if self.args.debug:
            debug = True
        verbose = False
        if self.args.verbose:
            verbose = True
        self.app = Application(debug=debug, db_path=self.args.database_dir, verbose=verbose)

    def _do_build_interface(self):
        self.build_parser = self.subparsers.add_parser(
            name="build",
            epilog="Please use '--' to separate options and arguments.",
            help="Build container images using the given Ansible playbook"
        )
        self.build_parser.add_argument(
            "playbook_path", metavar="PLAYBOOK_PATH",
            help="path to Ansible playbook"
        )
        self.build_parser.add_argument(
            "base_image", metavar="BASE_IMAGE",
            help="name of a container image to use as a base",
            nargs="?"
        )
        self.build_parser.add_argument(
            "target_image", metavar="TARGET_IMAGE",
            help="name of the built container image",
            nargs="?"
        )
        self.build_parser.add_argument("--builder", help="pick preferred builder backend",
                                       default="buildah",
                                       choices=["docker", "buildah"])
        self.build_parser.add_argument(
            "--no-cache",
            action="store_true",
            default=None,
            help="disable caching mechanism: storing layers and loading them if a task is unchanged; "
                 "this option also implies the final image is composed of a base image and one additional layer"
        )
        self.build_parser.add_argument(
            "--squash",
            action="store_true",
            default=False,
            help="squash final image down to a single layer"
        )
        self.build_parser.add_argument(
            "--build-volumes",
            help="mount selected directory inside the container during build, "
                 "should be specified as '/host/dir:/container/dir'",
            nargs="*"
        )
        self.build_parser.add_argument(
            "--build-user",
            help="the container gets invoked with this user during build"
        )
        self.build_parser.add_argument(
            "--inventory",
            help="path to Ansible inventory",
            default=None,
        )
        self.build_parser.add_argument(
            "--build-entrypoint",
            help="the container will be invoked with this entrypoint during the build"
        )
        self.build_parser.add_argument(
            "-w", "--workdir",
            help="path to an implicit working directory in the container"
        )
        self.build_parser.add_argument(
            "-l", "--label",
            help="add a label to the metadata of the image, "
                 "should be specified as 'key=value'",
            nargs="*",
            dest="labels"
        )
        self.build_parser.add_argument(
            "--annotation",
            help="Add key=value annotation for the target image",
            nargs="*",
            dest=ANNOTATIONS_KEY
        )
        # TODO: docker allows -e KEY and it is inherited from the current env
        self.build_parser.add_argument(
            "-e", "--env-vars",
            help="add an environment variable to the metadata of the image, "
                 "should be specified as 'KEY=VALUE'",
            nargs="*"
        )
        self.build_parser.add_argument(
            "--cmd",
            help="command to run by default in the container"
        )
        self.build_parser.add_argument(
            "--entrypoint",
            help="entrypoint script to configure for the container"
        )
        self.build_parser.add_argument(
            "-u", "--user",
            help="the container gets invoked with this user by default"
        )
        self.build_parser.add_argument(
            "-p", "--ports",
            help="ports to expose from container by default",
            nargs="*"
        )
        self.build_parser.add_argument(
            "--runtime-volumes",
            help="path a directory which has data stored outside of the container",
            nargs="*"
        )
        self.build_parser.add_argument(
            "--extra-buildah-from-args",
            help="arguments passed to buildah from command (be careful!)"
        )
        self.build_parser.add_argument(
            "--extra-buildah-run-args",
            help="arguments passed to buildah run command (be careful!)"
        )
        self.build_parser.add_argument(
            "--extra-podman-run-args",
            help="arguments passed to podman run command (be careful!)"
        )
        self.build_parser.add_argument(
            "--extra-ansible-args",
            help="arguments passed to ansible-playbook command (be careful!)"
        )
        self.build_parser.add_argument(
            "--python-interpreter",
            help="Path to a python interpreter inside the base image"
        )
        self.build_parser.set_defaults(subcommand="build")

        self.bio_parser = self.subparsers.add_parser(
            name="build-inside-openshift",
            description="Build image within an openshift environment.",
            help="Build image within an openshift environment."
        )
        self.bio_parser.set_defaults(subcommand="bio")

    def _do_get_logs_interface(self):
        self.gl_parser = self.subparsers.add_parser(
            name="get-logs",
            description="show logs of a specific build (default to latest build)",
            help="show logs of a specific build (default to latest build)"
        )
        self.gl_parser.add_argument(
            "BUILD_ID",
            help="ID of the build",
            nargs="?",
            default=None
        )
        self.gl_parser.set_defaults(subcommand="get-logs")

    def _do_list_builds_interface(self):
        self.lb_parser = self.subparsers.add_parser(
            name="list-builds",
            description="print a list of past and present builds",
            help="print a list of past and present builds"
        )
        self.lb_parser.set_defaults(subcommand="list-builds")

    def _do_inspect_interface(self):
        self.inspect_parser = self.subparsers.add_parser(
            name="inspect",
            description="provide detailed information for a selected build (default to latest build)",
            help="provide detailed information for a selected build (default to latest build)"
        )
        self.inspect_parser.add_argument(
            "BUILD_ID",
            help="ID of the build",
            nargs="?",
            default=None
        )
        self.inspect_parser.add_argument(
            "--json",
            help="output the information in JSON format",
            action="store_true"
        )
        self.inspect_parser.set_defaults(subcommand="inspect")

    def _do_push_interface(self):
        self.push_parser = self.subparsers.add_parser(
            name="push",
            description="push built image to a different location (default to latest build)",
            epilog="This command is thin wrapper on top of `podman push` command. "
                   "The target is passed directly to podman, for more information, please consult "
                   " podman-push(1) manpage or skopeo(1).",
            help="push built image to a different location (default to latest build)",
        )
        self.push_parser.add_argument(
            "TARGET",
            help="Target is composed of \"transport:details\"",
            default=None
        )
        self.push_parser.add_argument(
            "BUILD_ID",
            help="ID of the build",
            nargs="?",
            default=None
        )
        # nothing to force so far
        # self.push_parser.add_argument(
        #     "--force",
        #     help="Bypass any checks and just go for it",
        #     action="store_true"
        # )
        self.push_parser.set_defaults(subcommand="push")

    def _do_init_interface(self):
        self.lb_parser = self.subparsers.add_parser(
            name="init",
            description="Add a template playbook with all the vars.",
            help="Add a template playbook with all the vars."
        )
        self.lb_parser.set_defaults(subcommand="init")

    def _do_clean_interface(self):
        self.lb_parser = self.subparsers.add_parser(
            name="clean",
            description="Clean images from database which are no longer present on the disk",
            help="Clean images from database which are no longer present on the disk"
        )
        self.lb_parser.set_defaults(subcommand="clean")

    def _build(self):
        pb_vars_p = AnsibleVarsParser(self.args.playbook_path, self.args.inventory)
        build, metadata = pb_vars_p.get_build_and_metadata()
        build.metadata = metadata
        if self.args.workdir:
            metadata.working_dir = self.args.workdir
        if self.args.labels:
            for label in self.args.labels:
                err_msg = "Label variable {} doesn't seem to be " + \
                          "specified in format 'KEY=VALUE'.".format(label)
                k, v = split_once_or_fail_with(label, "=", err_msg)
                metadata.labels[k] = v
        if self.args.annotations:
            for ann in self.args.annotations:
                err_msg = "Annotation {} doesn't seem to be " + \
                          "specified in format 'KEY=VALUE'.".format(ann)
                k, v = split_once_or_fail_with(ann, "=", err_msg)
                metadata.annotations[k] = v
        if self.args.env_vars:
            for e_v in self.args.env_vars:
                err_msg = "Environment variable {} doesn't seem to be " + \
                    "specified in format 'KEY=VALUE'.".format(e_v)
                k, v = split_once_or_fail_with(e_v, "=", err_msg)
                metadata.env_vars[k] = v
        if self.args.cmd:
            metadata.cmd = self.args.cmd
        if self.args.entrypoint:
            metadata.entrypoint = self.args.entrypoint
        if self.args.user:
            metadata.user = self.args.user
        if self.args.ports:
            metadata.ports = self.args.ports
        if self.args.runtime_volumes:
            metadata.volumes = self.args.runtime_volumes

        build.playbook_path = self.args.playbook_path
        if self.args.build_volumes:
            build.build_volumes = self.args.build_volumes
        if self.args.base_image:
            build.base_image = self.args.base_image
        if self.args.target_image:
            build.target_image = self.args.target_image
        if self.args.builder:
            build.builder_name = self.args.builder
        if self.args.no_cache is not None:
            build.cache_tasks = not self.args.no_cache
        if self.args.squash:
            build.squash = self.args.squash
        if self.args.extra_buildah_from_args:
            build.buildah_from_extra_args = self.args.extra_buildah_from_args
        if self.args.extra_buildah_run_args:
            build.buildah_run_extra_args = self.args.extra_buildah_run_args
        if self.args.extra_podman_run_args:
            build.podman_run_extra_args = self.args.extra_podman_run_args
        if self.args.extra_ansible_args:
            build.ansible_extra_args = self.args.extra_ansible_args
        if self.args.python_interpreter:
            build.python_interpreter = self.args.python_interpreter
        if self.args.build_entrypoint:
            build.build_entrypoint = self.args.build_entrypoint

        self.app.build(build)

    def _build_inside_openshift(self):
        build_inside_openshift(self.app)

    def _list_builds(self):
        builds = self.app.list_builds()
        header = ("BUILD ID", "IMAGE NAME", "STATUS", "DATE", "BUILD TIME")
        builds_data = []
        for b in builds:
            build_time = ""
            if b.build_finished_time and b.build_start_time:
                build_time = fancy_time(b.build_finished_time - b.build_start_time)
            builds_data.append((
                b.build_id,
                b.target_image,
                b.state.value,
                b.build_finished_time if b.build_finished_time else "",
                build_time
            ))
        print(tabulate(builds_data, headers=header))

    def _get_logs(self):
        build_id = self.args.BUILD_ID
        log_lines = self.app.get_logs(build_id=build_id)
        if log_lines:
            print("\n".join(log_lines))
        else:
            print(f"There are no logs for build {build_id}")

    def _inspect(self):
        build_id = self.args.BUILD_ID
        inspect_data = self.app.inspect(build_id=build_id)
        if self.args.json:
            print(json.dumps(inspect_data))
        else:
            yaml.safe_dump(inspect_data, sys.stdout, indent=2, default_flow_style=False)

    def _push(self):
        build_id = self.args.BUILD_ID
        target = self.args.TARGET
        # force = self.args.force
        self.app.push(target, build_id=build_id, force=False)

    def _clean(self):
        builds = self.app.list_builds()
        print("Cleaning images from database which are no longer present on the disk...")
        for b in builds:
            try:
                run_cmd(["podman", "inspect", b.target_image],
                        return_output=False, log_output=False)
            except subprocess.CalledProcessError:
                self.app.remove_build(b.build_id)
                print(f"Build entry with ID {b.build_id} has been removed from DB as it no longer has it's corresponding image")
                continue
        print("Done!")

    def _init(self):
        with open('playbook.yml', 'w') as fd:
            fd.write(PLAYBOOK_TEMPLATE)
        print("Created an Ansible playbook template as playbook.yml")
        print("Edit the playbook accordingly and run the build subcommand:\n\nansible-bender build playbook.yaml")

    def run(self):
        subcommand = getattr(self.args, "subcommand", "nope")
        try:
            if self.args.version:
                try:
                    print(pkg_resources.get_distribution("ansible-bender").version)
                except pkg_resources.DistributionNotFound:
                    print("Ansible-bender is not installed, therefore we can't print the version.")
                    print("Please run `python3 setup.py --version` instead.")
                    return 18
                return 0
            if subcommand == "build":
                self._build()
                return 0
            elif subcommand == "list-builds":
                self._list_builds()
                return 0
            elif subcommand == "get-logs":
                self._get_logs()
                return 0
            elif subcommand == "inspect":
                self._inspect()
                return 0
            elif subcommand == "push":
                self._push()
                return 0
            elif subcommand == "bio":
                self._build_inside_openshift()
                return 0
            elif subcommand == "clean":
                self._clean()
                return 0
            elif subcommand == "init":
                self._init()
                return 0
        except KeyboardInterrupt:
            return 133
        except Exception as ex:
            self.app.clean()
            stderr = getattr(ex, "stderr", "")
            if stderr:
                print(stderr, file=sys.stderr)
            if self.args.debug:
                raise
            print("There was an error during execution: %s" % ex, file=sys.stderr)
            return 2
        self.parser.print_help()
        return 1


def main():
    sys.exit(CLI().run())


if __name__ == '__main__':
    main()



"""
ab (ansible builder) is a tool which enables you to build container images using ansible

## Usage

$ ab build ./playbook.yaml fedora:28 my-custom-image

"""
from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass



import datetime
import json
import logging
import os
import re
import shlex
import subprocess
import tempfile
from pathlib import Path
from typing import Optional, List

from ansible_bender.builders.base import Builder
from ansible_bender.constants import TIMESTAMP_FORMAT_TOGETHER
from ansible_bender.utils import graceful_get, run_cmd, buildah_command_exists, \
    podman_command_exists


logger = logging.getLogger(__name__)


def inspect_resource(resource_type, resource_id):
    try:
        i = run_cmd(["buildah", "inspect", "-t", resource_type, resource_id],
                    return_output=True, log_output=False)
    except subprocess.CalledProcessError:
        logger.info("no such %s %s", resource_type, resource_id)
        return None
    try:
        metadata = json.loads(i)
    except IndexError:
        logger.info("no such %s %s", resource_type, resource_id)
        return None
    return metadata


def get_buildah_image_id(container_image):
    metadata = inspect_resource("image", container_image)
    return graceful_get(metadata, "FromImageID")


def pull_buildah_image(container_image):
    run_cmd(["buildah", "pull", "--quiet", container_image],
            save_output_in_exc=False,
            log_stderr=False, print_output=True, log_output=False)


def does_image_exist(container_image):
    # cmd = ["podman", "image", "exists", container_image]
    # https://github.com/containers/libpod/issues/2924
    # https://github.com/ansible-community/ansible-bender/issues/114
    cmd = ["buildah", "inspect", "-t", "image", container_image]
    run_cmd(cmd, print_output=False)


def podman_run_cmd(container_image, cmd, extra_args, log_stderr=True, return_output=False):
    """
    run provided command in selected container image using podman; raise exc when command fails

    :param container_image: str
    :param cmd: list of str
    :param log_stderr: bool, log errors to stdout as ERROR level
    :param return_output: bool, if True, return output of the command
    :return: stdout output
    """
    return run_cmd(["podman", "run", "--rm"] + extra_args + [container_image] + cmd,
                   return_output=return_output, log_stderr=log_stderr)


def buildah_run_cmd(
        container_image: str,
        host_name: str,
        cmd: List[str],
        log_stderr: bool = True,
        log_output: bool = False,
        extra_from_args: Optional[str] = None,
        extra_run_args: Optional[List[str]] = None,
    ):
    """
    run provided command in selected container image using buildah; raise exc when command fails

    :param container_image: str
    :param host_name: str
    :param cmd: list of str
    :param log_stderr: bool, log errors to stdout as ERROR level
    :param log_output: bool, print output of the command to logs
    :param extra_from_args: a list of extra arguments for `buildah from`
    :param extra_run_args: a list of extra arguments for `buildah run`
    """
    extra_run_args = extra_run_args if extra_run_args is not None else []
    container_name = "{}-{}".format(host_name, datetime.datetime.now().strftime(TIMESTAMP_FORMAT_TOGETHER))
    # was the temporary container created? if so, remove it
    created = False
    try:
        create_buildah_container(
            container_image, container_name, build_volumes=None, extra_from_args=extra_from_args, debug=False)
        created = True
        run_cmd(["buildah", "run"] + extra_run_args + [container_name] + cmd, log_stderr=log_stderr, log_output=log_output)
    except subprocess.CalledProcessError:
        logger.error(f"Unable to create or run a container using {container_image} with buildah")
        raise
    finally:
        if created:
            run_cmd(["buildah", "rm", container_name], log_stderr=log_stderr)


def create_buildah_container(container_image, container_name, build_volumes=None, extra_from_args=None, debug=False):
    """
    Create new buildah container according to spec.

    :param container_image: name of the image
    :param container_name: name of the container to work in
    :param extra_from_args: a list of extra arguments for `buildah from`
    :param build_volumes: list of str, bind-mount specification: ["/host:/cont", ...]
    :param debug: bool, make buildah print debug info
    """
    args = []
    if build_volumes:
        for volume in build_volumes:
            args += ["-v", volume]
    if not extra_from_args is None:
        args += shlex.split(extra_from_args)
    args += ["--name", container_name, container_image]
    # will pull the image by default if it's not present in buildah's storage
    buildah("from", args, debug=debug, log_stderr=True)


def configure_buildah_container(container_name, working_dir=None, env_vars=None,
                                labels=None, annotations=None,
                                user=None, cmd=None, entrypoint=None,
                                ports=None, volumes=None,
                                debug=False):
    """
    apply metadata on the container so they get inherited in an image

    :param container_name: name of the container to work in
    :param working_dir: str, path to a working directory within container image
    :param labels: dict with labels
    :param annotations: dict with annotations
    :param env_vars: dict with env vars
    :param cmd: str, command to run by default in the container
    :param entrypoint: str, entrypoint script to configure for the container
    :param user: str, username or uid; the container gets invoked with this user by default
    :param ports: list of str, ports to expose from container by default
    :param volumes: list of str; paths within the container which has data stored outside
                    of the container
    :param debug: bool, make buildah print debug info
    """
    config_args = []
    if working_dir:
        config_args += ["--workingdir", working_dir]
    if env_vars:
        for k, v in env_vars.items():
            config_args += ["-e", "%s=%s" % (k, v)]
    if labels:
        for k, v in labels.items():
            config_args += ["-l", "%s=%s" % (k, v)]
    if annotations:
        for k, v in annotations.items():
            config_args += ["--annotation", "%s=%s" % (k, v)]
    if user:
        if os.getuid() != 0:
            logger.warning(
                "CHANGING USER IN ROOTLESS MODE IS DISCOURAGED, "
                "USE IT ONLY WHEN YOU KNOW WHAT YOU'RE DOING, "
                "YOU'VE BEEN WARNED"
            )
        config_args += ["--user", user]
    if cmd:
        config_args += ["--cmd", cmd]
    if entrypoint:
        config_args += ["--entrypoint", entrypoint]
    if ports:
        for p in ports:
            config_args += ["-p", p]
    if volumes:
        for v in volumes:
            config_args += ["-v", v]
    if config_args:
        buildah("config", config_args + [container_name], debug=debug)
    return container_name


def buildah(command, args_and_opts, print_output=False, debug=False, log_stderr=False):
    cmd = ["buildah"]
    # if debug:
    #     cmd += ["--debug"]
    cmd += [command] + args_and_opts
    logger.debug("running command: %s", command)
    return run_cmd(cmd, print_output=print_output, log_stderr=log_stderr)


def buildah_with_output(command, args_and_opts, debug=False):
    cmd = ["buildah"]
    # if debug:
    #     cmd += ["--debug"]
    cmd += [command] + args_and_opts
    output = run_cmd(cmd, return_output=True)
    logger.debug("output: %s", output)
    return output


class BuildahBuilder(Builder):
    ansible_connection = "buildah"
    name = "buildah"

    def __init__(self, build, debug=False):
        """
        :param build: instance of Build
        :param debug: bool, run buildah in debug or not?
        """
        super().__init__(build, debug=debug)
        self.target_image = build.target_image
        self.ansible_host = build.build_container
        self.logs = []
        buildah_command_exists()
        podman_command_exists()
        self.podman_run_args = []
        if self.build.podman_run_extra_args:
          self.podman_run_args = shlex.split(self.build.podman_run_extra_args)
        self.buildah_run_args = []
        if self.build.buildah_run_extra_args:
          self.buildah_run_args = shlex.split(self.build.buildah_run_extra_args)

    def create(self):
        """
        create a container where all the work happens
        """
        create_buildah_container(
            self.build.get_top_layer_id(), self.ansible_host,
            build_volumes=self.build.build_volumes,
            extra_from_args=self.build.buildah_from_extra_args,
            debug=self.debug)
        # let's apply configuration before execing the playbook, except for user
        configure_buildah_container(
            self.ansible_host, working_dir=self.build.metadata.working_dir,
            user=self.build.build_user,
            env_vars=self.build.metadata.env_vars,
            ports=self.build.metadata.ports,
            labels=self.build.metadata.labels,  # labels are not applied when they are configured
                                                # before doing commit
            annotations=self.build.metadata.annotations,
            entrypoint=self.build.build_entrypoint,
            debug=self.debug
        )

    def run(self, image_name, command):
        """
        run provided command in the selected image and return output

        :param image_name: str
        :param command: list of str
        :return: str (output)
        """
        cmd = ["podman", "run", "--rm"] + self.podman_run_args + [image_name] + command
        return run_cmd(cmd, return_output=True)

    def swap_working_container(self):
        """
        remove current working container and replace it with the provided one
        """
        self.clean()
        self.create()

    def commit(self, image_name: Optional[str] = None, print_output: bool = True, final_image: bool = False):
        """
        commit container into an image

        :param image_name: name of the image
        :param print_output: print to stdout if True
        :param final_image: is this is the final layer?
        :return:
        """
        if final_image:
            user = self.build.metadata.user
        else:
            user = self.build.build_user

        if (self.build.metadata.user or self.build.metadata.cmd or
            self.build.metadata.entrypoint or self.build.metadata.volumes):
            # change user if needed
            configure_buildah_container(
                self.ansible_host,
                user=user,
                cmd=self.build.metadata.cmd,
                entrypoint=self.build.metadata.entrypoint,
                volumes=self.build.metadata.volumes,
            )

        if image_name:
            args = [self.ansible_host, image_name]
            if final_image and self.build.squash:
                args.insert(0, "--squash")
            buildah("commit", args, print_output=print_output, debug=self.debug)
            return self.get_image_id(image_name)
        else:
            fd, name = tempfile.mkstemp()
            os.close(fd)
            args = ["-q", "--iidfile", name, self.ansible_host]
            # buildah 1.7.3 dropped the requirement for image name, let's support both
            # https://github.com/ansible-community/ansible-bender/issues/166
            if self.get_buildah_version() < (1, 7, 3):
                args += ["{}-{}".format(
                    self.ansible_host,
                    datetime.datetime.now().strftime(TIMESTAMP_FORMAT_TOGETHER)
                )]
            if final_image and self.build.squash:
                args.insert(0, "--squash")
            try:
                buildah("commit", args, print_output=print_output, debug=self.debug)
                image_id = Path(name).read_text()
                logger.debug("layer id = %s", image_id)
                return image_id
            finally:
                os.unlink(name)

    def clean(self):
        """
        clean working container
        """
        buildah("rm", [self.ansible_host], debug=self.debug)

    def get_image_id(self, image_name):
        """ return image_id for provided image """
        image_id = get_buildah_image_id(image_name)
        if not image_id:
            raise RuntimeError("We haven't got any image ID: the image is not present "
                               "or buildah/podman is malfunctioning.")
        return image_id

    def is_image_present(self, image_reference):
        """
        :return: True when the selected image is present, False otherwise
        """
        if not image_reference:
            return False
        try:
            does_image_exist(image_reference)
        except subprocess.CalledProcessError:
            return False
        else:
            return True

    def pull(self):
        """
        pull base image
        """
        logger.info("pulling base image: %s", self.build.base_image)
        pull_buildah_image(self.build.base_image)

    def push(self, build, target, force=False):
        """
        push built image into a remote location using `podman push`

        :param target: str, transport:details
        :param build: instance of Build
        :param force: bool, bypass checks if True
        :return: None
        """
        built_image = build.get_target_image_id()
        cmd = ["buildah", "push", built_image, target]
        # podman prints progress to stderr
        run_cmd(cmd, print_output=False, log_stderr=False)

    def find_python_interpreter(self):
        """
        find python executable in the base image

        :return: str, path to python interpreter
        """
        for i in self.python_interpr_prio:
            cmd = ["ls", i]
            try:
                podman_run_cmd(self.build.base_image, cmd, self.podman_run_args, log_stderr=False)
            except subprocess.CalledProcessError:
                logger.info("python interpreter %s does not exist", i)
                continue
            else:
                logger.info("using python interpreter %s", i)
                return i
        logger.error("couldn't locate python interpreter, tried these paths: %s", self.python_interpr_prio)
        raise RuntimeError(f"no python interpreter was found in the base image \"{self.build.base_image}\""
                           ", you can specify the path via CLI option --python-interpreter")

    def get_logs(self):
        """
        obtain logs for the selected build

        :return: list of str
        """

    def sanity_check(self):
        """
        invoke container tooling and thus verify they work well
        """
        # doing podman info would be super-handy, but it will immensely clutter the output
        logger.debug("checking that podman command works")
        run_cmd(["podman", "version"], log_stderr=True, log_output=True)
        logger.debug("checking that buildah command works")
        run_cmd(["buildah", "version"], log_stderr=True, log_output=True)
        logger.debug("Checking container creation using buildah")
        buildah_run_cmd(
            self.build.base_image, self.ansible_host, ["true"],
            log_stderr=True, extra_from_args=self.build.buildah_from_extra_args, extra_run_args=self.buildah_run_args)

    def get_buildah_version(self):
        out = run_cmd(["buildah", "version"], log_stderr=True, return_output=True, log_output=False)
        version = re.findall(r"Version:\s*([\d\.]+)", out)[0].split(".")
        logger.debug("buildah version = %s", version)
        # buildah version = ['1', '11', '3']
        try:
            return tuple(map(int, version))
        except (IndexError, ValueError) as ex:
            logger.error("Unable to parse buildah's version: %s", ex)
            return 0, 0, 0

    def check_container_creation(self):
        """
        check that containers can be created
        """
        logger.debug("trying to create a dummy container using podman")
        podman_run_cmd(self.build.base_image, ["true"], self.podman_run_args, log_stderr=True)



"""
Builder interface
"""
import logging

from ansible_bender.builders.buildah_builder import BuildahBuilder


logger = logging.getLogger(__name__)


BUILDERS = {
    BuildahBuilder.name: BuildahBuilder
}


def get_builder(builder_name):
    try:
        return BUILDERS[builder_name]
    except KeyError:
        raise RuntimeError("No such builder %s" % builder_name)



"""
This is a configuration module
"""
import datetime

from ansible_bender.builders.base import BuildState
from ansible_bender.constants import TIMESTAMP_FORMAT, ANNOTATIONS_KEY
from ansible_bender.exceptions import ABValidationError
from ansible_bender.schema import IMAGE_META_SCHEMA, BUILD_SCHEMA 
from ansible_bender.utils import graceful_get

import jsonschema


class ImageMetadata:
    """
    working_dir: str,
    labels: dict with labels
    env_vars: dict with env vars
    cmd: str, command to run by default in the container
    entrypoint: str, entrypoint script to configure for the container
    user: str, username or uid; the container gets invoked with this user by default
    ports: list of str, ports to expose from container by default
    volumes: list of str; paths within the container which has data stored outside
             of the container
    """
    def __init__(self):
        self.working_dir = None
        self.labels = {}
        self.annotations = {}
        self.env_vars = {}
        self.cmd = None
        self.entrypoint = None
        self.user = None
        self.ports = []
        self.volumes = []

    def to_dict(self):
        return {
            "working_dir": self.working_dir,
            "labels": self.labels,
            ANNOTATIONS_KEY: self.annotations,
            "env_vars": self.env_vars,
            "cmd": self.cmd,
            "entrypoint": self.entrypoint,
            "user": self.user,
            "ports": self.ports,
            "volumes": self.volumes
        }

    def update_from_configuration(self, data):
        """ update current object with data provided from Ansible vars """
        if not isinstance(data, dict):
            raise ABValidationError(f"\"target_image\" should be dict, not {data.__class__.__name__}")
        self.working_dir = data.get("working_dir", None)
        self.labels.update(data.get("labels", {}))
        self.annotations.update(data.get(ANNOTATIONS_KEY, {}))
        self.env_vars.update(data.get("environment", {}))
        self.cmd = data.get("cmd", None)
        self.entrypoint = data.get("entrypoint", None)
        self.user = data.get("user", None)
        self.ports += data.get("ports", [])
        self.volumes += data.get("volumes", [])

    @classmethod
    def from_json(cls, j):
        """ return Metadata instance from the provided json """
        m = cls()
        m.working_dir = j["working_dir"]
        m.labels = j["labels"]
        m.annotations = graceful_get(j, ANNOTATIONS_KEY, default={})
        m.env_vars = j["env_vars"]
        m.cmd = j["cmd"]
        m.entrypoint = j.get("entrypoint", "")
        m.user = j["user"]
        m.ports = j["ports"]
        m.volumes = j["volumes"]
        return m

    def validate(self):
        jsonschema.validate(self.to_dict(), IMAGE_META_SCHEMA)


class Layer:
    """ This is an image layer """

    def __init__(self, content, layer_id, base_image_id, cached=None):
        """
        :param content: what's the content of the layer
        :param layer_id:
        :param base_image_id:
        :param cached: bool, was it loaded from cache?
        """
        self.content = content
        self.layer_id = layer_id
        self.base_image_id = base_image_id
        self.cached = cached

    def __str__(self):
        return f"layer_id={self.layer_id} cached={self.cached}"

    def __repr__(self):
        return f"Layer(layer_id={self.layer_id}, cached={self.cached}, base={self.base_image_id})"

    def to_dict(self):
        return {
            "content": self.content,
            "layer_id": self.layer_id,
            "base_image_id": self.base_image_id,
            "cached": self.cached
        }

    @classmethod
    def from_json(cls, j):
        return cls(
            j["content"],
            j["layer_id"],
            j["base_image_id"],
            cached=j["cached"]
        )


class Build:
    """ class which represents a build """
    def __init__(self):
        self.build_id = None  # PK, should be set by database
        self.playbook_path = None
        self.build_volumes = []  # volumes for the build container
        self.build_user = None
        self.build_entrypoint = None
        self.metadata = None  # Image metadata
        self.state = BuildState.NEW
        self.build_start_time = None
        self.build_finished_time = None
        self.base_image = None
        self.build_container = None
        self.target_image = None
        self.builder_name = None
        self.layers = []
        self.final_layer_id = None  # once the image is built, this is the final layer: content + metadata
        self.layer_index = {}  # this is an index for layers: `layer_id: Layer()`
        self.cache_tasks = True  # we cache by default, a user can opt out
        self.log_lines = []  # a list of strings
        self.layering = True
        self.squash = False
        self.debug = False
        self.verbose = False
        self.pulled = False  # was the base image pulled?
        self.buildah_from_extra_args = None
        self.buildah_run_extra_args = None
        self.podman_run_extra_args = None
        self.ansible_extra_args = None
        self.python_interpreter = None
        self.verbose_layer_names = False

    def to_dict(self):
        """ serialize """
        return {
            "build_id": self.build_id,
            "playbook_path": self.playbook_path,
            "build_volumes": self.build_volumes,
            "build_user": self.build_user,
            "metadata": self.metadata.to_dict(),
            "state": self.state.value,
            "build_start_time": self.build_start_time.strftime(TIMESTAMP_FORMAT)
                if self.build_start_time else None,
            "build_finished_time": self.build_finished_time.strftime(TIMESTAMP_FORMAT)
                if self.build_finished_time else None,
            "base_image": self.base_image,
            "target_image": self.target_image,
            "builder_name": self.builder_name,
            "layers": [x.to_dict() for x in self.layers],
            "final_layer_id": self.final_layer_id,
            "layer_index": {x.layer_id: x.to_dict() for x in self.layers},
            "build_container": self.build_container,
            "cache_tasks": self.cache_tasks,
            # we could compress/base64 here, let's go for the easier solution first
            "log_lines": self.log_lines,
            "layering": self.layering,
            "squash": self.squash,
            "debug": self.debug,
            "verbose": self.verbose,
            "pulled": self.pulled,
            "buildah_from_extra_args": self.buildah_from_extra_args,
            "buildah_run_extra_args": self.buildah_run_extra_args,
            "podman_run_extra_args": self.podman_run_extra_args,
            "ansible_extra_args": self.ansible_extra_args,
            "python_interpreter": self.python_interpreter,
            "verbose_layer_names": self.verbose_layer_names,
        }

    def update_from_configuration(self, data):
        """ update current object with data provided from Ansible vars """
        self.build_volumes += graceful_get(data, "working_container", "volumes", default=[])
        self.build_user = graceful_get(data, "working_container", "user")
        self.build_entrypoint = graceful_get(data, "working_container", "entrypoint")
        self.base_image = graceful_get(data, "base_image")
        self.target_image = graceful_get(data, "target_image", "name")
        # self.builder_name = None
        self.cache_tasks = graceful_get(data, "cache_tasks", default=self.cache_tasks)
        self.layering = graceful_get(data, "layering", default=self.layering)
        self.squash = graceful_get(data, "squash", default=self.squash)
        self.buildah_from_extra_args = graceful_get(data, "buildah_from_extra_args")
        self.buildah_run_extra_args = graceful_get(data, "buildah_run_extra_args")
        self.podman_run_extra_args = graceful_get(data, "podman_run_extra_args")
        self.ansible_extra_args = graceful_get(data, "ansible_extra_args")
        self.verbose_layer_names = graceful_get(data, "verbose_layer_names")
        # we should probably get this from the official Ansible variable
        # self.python_interpreter = None

    @classmethod
    def from_json(cls, j):
        """ return Build instance from the provided json """
        b = cls()
        b.build_id = j["build_id"]
        b.playbook_path = j.get("playbook_path", None)
        b.build_volumes = j["build_volumes"]
        b.build_user = graceful_get(j, "build_user")
        b.build_entrypoint = graceful_get(j, "build_entrypoint")
        b.metadata = ImageMetadata.from_json(j["metadata"])
        b.state = BuildState(j["state"])
        b.build_start_time = None
        sta = j["build_start_time"]
        if sta:
            b.build_start_time = datetime.datetime.strptime(sta, TIMESTAMP_FORMAT)
        b.build_finished_time = None
        fin = j["build_finished_time"]
        if fin:
            b.build_finished_time = datetime.datetime.strptime(fin, TIMESTAMP_FORMAT)
        b.base_image = j["base_image"]
        b.target_image = j["target_image"]
        b.builder_name = j["builder_name"]
        b.layers = [Layer.from_json(x) for x in j["layers"]]
        b.final_layer_id = j.get("final_layer_id", None)
        b.layer_index = {layer_id: Layer.from_json(layer_data)
                         for layer_id, layer_data in j["layer_index"].items()}
        b.build_container = j["build_container"]
        b.cache_tasks = j["cache_tasks"]
        b.log_lines = j["log_lines"]
        b.layering = j["layering"]
        b.squash = j.get("squash", False)
        b.debug = j["debug"]
        b.verbose = j["verbose"]
        b.pulled = j["pulled"]
        b.buildah_from_extra_args = j.get("buildah_from_extra_args", None)
        b.ansible_extra_args = j.get("ansible_extra_args", None)
        b.buildah_run_extra_args = j.get("buildah_run_extra_args", None)
        b.podman_run_extra_args = j.get("podman_run_extra_args", None)
        b.python_interpreter = j.get("python_interpreter", None)
        b.verbose_layer_names = graceful_get(j, "verbose_layer_names", default=False)
        return b

    def record_layer(self, content, layer_id, base_image_id, cached=None):
        """
        record a new layer for this build

        :param content:
        :param layer_id:
        :param base_image_id:
        :param cached: bool, was it cached?
        """
        layer = Layer(content, layer_id, base_image_id, cached=cached)
        self.layers.append(layer)
        self.layer_index[layer_id] = layer

    def wipe_layers(self):
        """ remove all layers from the DB: used by squash """
        self.layers = []
        self.layer_index = {}

    def get_top_layer_id(self):
        """
        return id of the top layer, or None
        """
        if self.layers:
            return self.layers[-1].layer_id

    def get_target_image_id(self):
        """
        this is the preferred way of getting ID of the built image

        :return: str
        """
        if self.state not in [BuildState.DONE, BuildState.FAILED]:
            raise RuntimeError(f"Build have not finished yet, it is in state '{self.state.value}'.")
        return self.final_layer_id

    def was_last_layer_cached(self):
        if self.layers:
            return self.layers[-1].cached

    def record_cache_entry(self, image_id):
        self.layer_index[image_id].cached = True

    def stop_layering(self):
        self.layering = False

    def is_layering_on(self):
        return self.layering

    def validate(self):
        jsonschema.validate(self.to_dict(), BUILD_SCHEMA)

    def is_failed(self):
        return self.state == BuildState.FAILED



"""
A database module. A class to manage ab's persistent data.


# The schema

{
    "next_build_id": int  #
    "builds": {
        <id>: {
            state: ...
            base_image: ...
            target_image: ...
            builder_name: ...
            metadata: {
                command: ...
                user: ...
                env: ...
                ...
            },
            layers: [
                Layer(content, layer_id, base_image_id, cached),
            ]
            layer_index: {
                layer_id: layer
            }
        },
    },
    "store": {
        "base-image-id": {  # base-image + content = new image
            content: {
                image_id:
            }
        }
    }
}
"""
import copy
import datetime
import json
import logging
import os
import time
from contextlib import contextmanager

from ansible_bender.conf import Build
from ansible_bender.constants import TIMESTAMP_FORMAT

DEFAULT_DATA = {
    "next_build_id": 1,
    "builds": {},
    "store": {}
}

PATH_CANDIDATES = [
    "~/.cache",
    os.environ.get("XDG_RUNTIME_DIR", ""),
    "/var/tmp"
]

logger = logging.getLogger(__name__)


def generate_working_cont_name(image_name):
    timestamp = datetime.datetime.now().strftime(TIMESTAMP_FORMAT)
    san = image_name.replace(".", "-").replace(":", "-").replace("/", "-")
    return f"{san}-{timestamp}-cont"


class Database:
    """ Simple implementation of persistent data store for ab; it's just a locked json file """

    def __init__(self, db_path=None):
        path_preference = PATH_CANDIDATES.copy()
        if db_path:
            path_preference.insert(0, db_path)
        self.runtime_dir_path, self.db_root_path = self._runtime_dir_path(path_preference)

    @contextmanager
    def acquire(self):
        """
        lock usage of database
        """
        while True:
            try:
                with open(self._lock_path(), "r") as fd:
                    # the file exists, ab changes the database
                    pid = fd.read()
                logger.info("ab is running as PID %s", pid)
                # logger.debug("stack trace: %s", traceback.extract_stack())
                time.sleep(0.1)
            except FileNotFoundError:
                # cool, let's take the lock
                try:
                    with os.fdopen(os.open(self._lock_path(), os.O_WRONLY|os.O_EXCL|os.O_CREAT ),'w') as fd:
                        fd.write("%s" % os.getpid())
                    break
                except FileExistsError:
                    continue
        # logger.debug("this stack has the lock: %s", traceback.extract_stack())
        yield True
        self.release()

    def release(self):
        """ release lock """
        try:
            os.unlink(self._lock_path())
        except FileNotFoundError:
            pass

    @staticmethod
    def _runtime_dir_path(path_preference):
        logger.debug("search for runtime dir")
        for c in path_preference:
            logger.debug("trying %s", c)
            if not c:
                continue
            resolved = os.path.abspath(os.path.expanduser(c))
            if os.path.isdir(resolved):
                break
        else:
            raise RuntimeError("Can't find a suitable directory to store runtime data.")
        logger.debug("runtime dir is %s", resolved)
        our_dir = os.path.join(resolved, "ab")
        os.makedirs(our_dir, mode=0o0700, exist_ok=True)
        return our_dir, resolved

    def _db_path(self):
        data_path = os.path.join(self.runtime_dir_path, "db.json")
        return data_path

    def _lock_path(self):
        lock_path = os.path.join(self.runtime_dir_path, "ab.pid")
        return lock_path

    def _load(self):
        """ load data from disk, lock has to be acquired already! """
        try:
            with open(self._db_path(), "r") as fd:
                return json.load(fd)
        except FileNotFoundError:
            # no problem, probably a first run
            logger.debug("initializing database")
            return copy.deepcopy(DEFAULT_DATA)

    @staticmethod
    def _load_build(data, build_id, is_latest=False):
        """
        load selected build from database
        :param data: dict
        :param build_id: str or None
        :param is_latest: bool
        :return: build
        """
        try:
            return Build.from_json(data["builds"][build_id])
        except KeyError:
            if is_latest:
                raise RuntimeError("Latest build with ID %s is no longer available, probably got cleaned." % build_id)
            else:
                raise RuntimeError("There is no such build with ID %s" % build_id)

    def _save(self, data):
        """ save data from memory to disk, lock has to be acquired already! """
        with open(self._db_path(), "w") as fd:
            json.dump(data, fd, indent=2)

    @staticmethod
    def _get_and_bump_build_id(data):
        """ return id for next build id and increment the one in DB """
        next_build_id = data["next_build_id"]
        data["next_build_id"] += 1
        if not data["builds"].get(str(next_build_id)):
            return str(next_build_id)
        else:
            raise Exception(f'Database seems to be corrupted. Build {next_build_id} already exists.')

    def record_build(self, build_i, build_id=None, build_state=None, set_finish_time=False):
        """
        record build into database

        :param build_i: Build instance
        :param build_id: str, id of the build to load from DB
        :param build_state: one of BuildState
        :param set_finish_time: bool, set build_finish_time to current time
        """
        with self.acquire():
            data = self._load()
            if build_id is not None:
                build_i = self._load_build(data, build_id)
            if build_state is not None:
                build_i.state = build_state
            if build_i.build_id is None:
                build_i.build_container = generate_working_cont_name(build_i.target_image)
                build_i.build_id = self._get_and_bump_build_id(data)
            if set_finish_time:
                build_i.build_finished_time = datetime.datetime.now()
            data["builds"][build_i.build_id] = build_i.to_dict()
            self._save(data)
        return build_i

    def get_latest_build(self):
        """
        return build with highest ID

        :return: build
        """
        with self.acquire():
            data = self._load()
            build_id = str(data["next_build_id"] - 1)
            return self._load_build(data, build_id, is_latest=True)

    def get_build(self, build_id):
        """
        get Build instance by selected build_id

        :param build_id: str
        :return: instance of Build
        """
        with self.acquire():
            data = self._load()
            return self._load_build(data, build_id)

    def save_layer(self, layer_id, base_image, content):
        with self.acquire():
            data = self._load()
            store = data["store"]
            store.setdefault(base_image, {})
            store[base_image].setdefault(content, {})
            store[base_image][content]["image_id"] = layer_id
            self._save(data)

    def get_cached_layer(self, content, base_image_id):
        with self.acquire():
            data = self._load()
            store = data["store"]
            try:
                return store[base_image_id][content]["image_id"]
            except KeyError:
                return

    def load_builds(self):
        """
        provide a list of all available builds

        :return: a list of Build instances
        """
        with self.acquire():
            data = self._load()
            return [Build.from_json(b) for b in data["builds"].values()]

    def delete_build(self, build_id):
        """
        delete a build from database

        :param build_id: str, id of the build to be deleted from DB
        """
        with self.acquire():
            data = self._load()
            try:
                del data["builds"][build_id]
            except KeyError:
                raise RuntimeError("There is no such build with ID %s" % build_id)
            finally:
                self._save(data)

    def load_python_interpreter(self, base_image_id):
        """
        loads the python interpreter path from the base image. Works for the top-level base image only, layer image ids will not be found.

        :param base_image_id: str, id/hash of the base image
        :return: python interpreter path if found, None otherwise
        """
        with self.acquire():
            data = self._load()
            store = data["store"]
            store.setdefault(base_image_id, {})
            return store[base_image_id].get("python_interpreter")

    def record_python_interpreter(self, base_image_id, python_interpreter):
        """
        records/caches the python interpreter path for later use.

        :param base_image_id: str, id/hash of the base image
        :param python_interpreter: str, path of the python interpreter on the base image
        """
        with self.acquire():
            data = self._load()
            store = data["store"]
            store.setdefault(base_image_id, {})
            store[base_image_id]["python_interpreter"] = python_interpreter
            self._save(data)
